"""Tests for the Pulse class."""
import itertools

import numpy as np
import pytest

from qililab.constants import PULSE
from qililab.pulse import Cosine, Drag, Gaussian, Pulse, Rectangular

# Parameters for the different Pulses
AMPLITUDE = [0, 0.9, 1.1, -0.8]
PHASE = [0, np.pi / 3, 2 * np.pi]
DURATION = [47, 40]
FREQUENCY = [0.7e9]
SHAPE = [
    Rectangular(),
    Cosine(),
    Cosine(lambda_2=0.3),
    Gaussian(num_sigmas=4),
    Drag(num_sigmas=4, drag_coefficient=1.0),
]

# Parameters of the envelope.
RESOLUTION = 1.0


@pytest.mark.parametrize(
    "pulse",
    [
        Pulse(amplitude=amplitude, phase=phase, duration=duration, frequency=frequency, pulse_shape=shape)
        for amplitude, phase, duration, frequency, shape in itertools.product(
            AMPLITUDE, PHASE, DURATION, FREQUENCY, SHAPE
        )
    ],
)
class TestPulse:
    """Unit tests checking the Pulse attributes and methods"""

    def test_envelope_method(self, pulse: Pulse):
        """Test envelope method"""
        envelope = pulse.envelope(resolution=RESOLUTION)

        # Test not None and type
        assert envelope is not None
        assert isinstance(envelope, np.ndarray)

        # Assert size of np.ndarray
        assert len(envelope) == pulse.duration / RESOLUTION

        # Test the maximums of the positive envelopes
        if pulse.amplitude >= 0.0:
            if isinstance(pulse.pulse_shape, Cosine) and pulse.pulse_shape.lambda_2 > 0.0:
                # If lambda_2 > 0.0 the max amplitude is reduced, but never gets down 70% of the Amplitude
                assert round(np.max(np.real(envelope)), 2) <= pulse.amplitude
                assert round(np.max(np.real(envelope)), 2) >= 0.7 * pulse.amplitude
            else:
                assert round(np.max(np.real(envelope)), 2) == pulse.amplitude

        # Test the minimums of the negative envelopes
        elif pulse.amplitude <= 0.0:
            if isinstance(pulse.pulse_shape, Cosine) and pulse.pulse_shape.lambda_2 > 0.0:
                # If lambda_2 > 0.0 the max amplitude is reduced, but never gets down 70% of the Amplitude
                assert round(np.min(np.real(envelope)), 2) >= pulse.amplitude
                assert round(np.min(np.real(envelope)), 2) <= 0.7 * pulse.amplitude
            else:
                assert round(np.min(np.real(envelope)), 2) == pulse.amplitude

        # Test the 0 amplitude case
        elif pulse.amplitude == 0.0:
            assert 0.0 == np.max(envelope) == np.min(envelope)

    def test_from_dict_method(self, pulse: Pulse):
        """Test to_dict method"""
        dictionary = pulse.to_dict()
        pulse2 = Pulse.from_dict(dictionary)

        dictionary2 = pulse2.to_dict()
        pulse3 = Pulse.from_dict(dictionary2)

        for p in [pulse2, pulse3]:
            assert p is not None
            assert isinstance(p, Pulse)

        assert pulse == pulse2 == pulse3

    def test_to_dict_method(self, pulse: Pulse):
        """Test to_dict method"""
        dictionary = pulse.to_dict()

        pulse2 = Pulse.from_dict(dictionary)
        dictionary2 = pulse2.to_dict()

        for dict_ in [dictionary, dictionary2]:
            assert dict_ is not None
            assert isinstance(dict_, dict)

        assert (
            dictionary
            == dictionary2
            == {
                PULSE.AMPLITUDE: pulse.amplitude,
                PULSE.FREQUENCY: pulse.frequency,
                PULSE.PHASE: pulse.phase,
                PULSE.DURATION: pulse.duration,
                PULSE.PULSE_SHAPE: pulse.pulse_shape.to_dict(),
            }
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from qililab.instruments.qblox.qblox_sequencer import QbloxSequencer
from qililab.typings import AcquireTriggerMode, IntegrationMode
from qililab.utils.castings import cast_enum_fields


@dataclass
class QbloxADCSequencer(QbloxSequencer):
    scope_acquire_trigger_mode: AcquireTriggerMode
    scope_hardware_averaging: bool
    sampling_rate: float  # default sampling rate for Qblox is 1.e+09
    hardware_demodulation: bool  # demodulation flag
    integration_length: int
    integration_mode: IntegrationMode
    sequence_timeout: int  # minutes
    acquisition_timeout: int  # minutes
    scope_store_enabled: bool
    threshold: float
    threshold_rotation: float
    time_of_flight: int  # nanoseconds

    def __post_init__(self):
        cast_enum_fields(obj=self)



"""This file tests the the connection class"""

import re

import pytest

from qililab.instrument_connections.connection import Connection


class TestConnection:
    """This class contains the unit tests for the ``Connection`` class."""

    def test_error_raises_when_already_connected(self):
        """Test that ensure that an error rises when connecting an instrument that is already connected"""
        test = Connection(settings={"address": "test"})
        test._connected = True
        expected = re.escape("Instrument (test) is already connected")
        with pytest.raises(ValueError, match=expected):
            test.connect(device="test", device_name="test")



"""This file tests the the ``qblox_d5a`` class"""

from unittest.mock import MagicMock

import pytest

from qililab.instruments.qblox import QbloxD5a
from qililab.typings.enums import Parameter


@pytest.fixture(name="qblox_d5a")
def fixture_qblox_d5a():
    """Fixture that returns an instance of a dummy QbloxD5a."""
    return QbloxD5a(
        {
            "alias": "d5a",
            "voltage": [0],
            "span": [],
            "ramping_enabled": [],
            "ramp_rate": [],
            "dacs": []
        }
    )


class TestQbloxD5a:
    """This class contains the unit tests for the ``qblox_d5a`` class."""

    def test_error_raises_when_no_channel_specified(self, qblox_d5a):
        """These test makes soure that an error raises whenever a channel is not specified in chainging a parameter

        Args:
            qblox_d5a (_type_): _description_
        """
        name = qblox_d5a.name.value
        with pytest.raises(ValueError, match=f"channel not specified to update instrument {name}"):
            qblox_d5a.device = MagicMock()
            qblox_d5a.set_parameter(parameter=Parameter.VOLTAGE, value="2", channel_id=None)

    def test_setup_method_no_connection(self, qblox_d5a):
        """Test setup method."""
        qblox_d5a.set_parameter(parameter=Parameter.VOLTAGE, value=2, channel_id=0)
        assert qblox_d5a.settings.voltage[0] == 2.0

    def test_initial_setup_method_no_connection(self, qblox_d5a):
        """Test setup method."""
        with pytest.raises(RuntimeError, match=f"Device of instrument {qblox_d5a.alias} has not been initialized."):
            qblox_d5a.initial_setup()
            qblox_d5a



import numpy as np
import pytest

from qililab.waveforms import Square


@pytest.fixture(name="square")
def fixture_square():
    return Square(amplitude=0.5, duration=100)


class TestSquare:
    def test_init(self, square):
        """Test __init__ method"""
        assert square.amplitude == 0.5
        assert square.duration == 100

    def test_envelope_method(self, square):
        """Test envelope method"""
        envelope = 0.5 * np.ones(round(100 / 1))
        assert np.allclose(square.envelope(), envelope)

    def test_get_duration_method(self, square):
        """Test get_duration method"""
        assert square.get_duration() == 100



"""Test the annealing program class"""

from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from tests.data import Galadriel

from qililab import AnnealingProgram
from qililab.qprogram.crosstalk_matrix import FluxVector
from qililab.settings.analog.flux_control_topology import FluxControlTopology


@pytest.fixture(name="anneal_program_dictionary")
def get_anneal_program_dictionary():
    """Dummy anneal program dictionary"""
    return [
        {
            "qubit_0": {"sigma_x": 0, "sigma_y": 0, "sigma_z": 1},
            "qubit_1": {"sigma_x": 0.1, "sigma_y": 0.1, "sigma_z": 0.1},
            "coupler_0_1": {"sigma_x": 1, "sigma_y": 0.2, "sigma_z": 0.2},
        },
        {
            "qubit_0": {"sigma_x": 0.1, "sigma_y": 0.1, "sigma_z": 1.1},
            "qubit_1": {"sigma_x": 0.2, "sigma_y": 0.2, "sigma_z": 0.2},
            "coupler_0_1": {"sigma_x": 0.9, "sigma_y": 0.1, "sigma_z": 0.1},
        },
        {
            "qubit_0": {"sigma_x": 0.3, "sigma_y": 0.3, "sigma_z": 0.7},
            "qubit_1": {"sigma_x": 0.5, "sigma_y": 0.2, "sigma_z": 0.01},
            "coupler_0_1": {"sigma_x": 0.5, "sigma_y": 0, "sigma_z": -1},
        },
    ]


@pytest.fixture(name="flux_to_bus_topology")
def get_flux_to_bus_topology():
    flux_control_topology_dict = [
        {"flux": "phix_q0", "bus": "flux_line_phix_q0"},
        {"flux": "phiz_q0", "bus": "flux_line_phiz_q0"},
        {"flux": "phix_q1", "bus": "flux_line_phix_q1"},
        {"flux": "phiz_q1", "bus": "flux_line_phiz_q1"},
        {"flux": "phix_c0_1", "bus": "flux_line_phix_c0_1"},
        {"flux": "phiz_c0_1", "bus": "flux_line_phiz_c0_1"},
    ]
    return [FluxControlTopology(**flux_control) for flux_control in flux_control_topology_dict]


@pytest.fixture(name="transpiled_program_dictionary")
def get_transpiled_program_dictionary():
    """Transpiled program for the anneal program above"""
    return [
        {
            "phix_q0": 0,
            "phiz_q0": 3,
            "phix_q1": 0.2,
            "phiz_q1": 0.30000000000000004,
            "phix_c0_1": 2,
            "phiz_c0_1": 0.6000000000000001,
        },
        {
            "phix_q0": 0.2,
            "phiz_q0": 3.3000000000000003,
            "phix_q1": 0.4,
            "phiz_q1": 0.6000000000000001,
            "phix_c0_1": 1.8,
            "phiz_c0_1": 0.30000000000000004,
        },
        {
            "phix_q0": 0.6,
            "phiz_q0": 2.0999999999999996,
            "phix_q1": 1.0,
            "phiz_q1": 0.03,
            "phix_c0_1": 1.0,
            "phiz_c0_1": -3,
        },
    ]


@pytest.fixture(name="annealing_program")
def get_annealing_program(flux_to_bus_topology, anneal_program_dictionary):
    """Build dummy annealing program"""
    return AnnealingProgram(flux_to_bus_topology=flux_to_bus_topology, annealing_program=anneal_program_dictionary)


@pytest.fixture(name="annealing_program_transpiled")
def get_annealing_program_transpiled(flux_to_bus_topology, anneal_program_dictionary, transpiled_program_dictionary):
    """Build dummy annealing program with fluxes already transpiled"""
    annealing_program = AnnealingProgram(
        flux_to_bus_topology=flux_to_bus_topology, annealing_program=anneal_program_dictionary
    )
    annealing_program._transpiled_program = transpiled_program_dictionary
    return annealing_program


def flux_to_bus(flux):
    """Return the corresponding bus to a given flux from the runcard topology"""
    return next(element["bus"] for element in Galadriel.runcard["flux_control_topology"] if element["flux"] == flux)


def dummy_transpiler(delta, epsilon):
    """Dummy transpiler for testing"""
    return (2 * delta, 3 * epsilon)


class TestAnnealingProgram:
    """Test class for the AnnealingProgram class"""

    def test_init(self, annealing_program, anneal_program_dictionary, flux_to_bus_topology):
        """Test init method"""
        assert annealing_program._flux_to_bus_topology == flux_to_bus_topology
        assert annealing_program._annealing_program == anneal_program_dictionary

    def test_transpile(self, annealing_program, transpiled_program_dictionary):
        """Test transpile method"""
        annealing_program.transpile(transpiler=dummy_transpiler)
        assert annealing_program._transpiled_program == transpiled_program_dictionary

    @pytest.mark.parametrize("minimum_clock_time", [1, 3, 4, 10])
    def test_get_waveforms(self, annealing_program_transpiled, minimum_clock_time):
        """Test get waveforms method works as intended"""
        anneal_waveforms = annealing_program_transpiled.get_waveforms(minimum_clock_time=minimum_clock_time)
        transpiled_program = annealing_program_transpiled._transpiled_program

        pad_length = (
            minimum_clock_time - len(transpiled_program) % minimum_clock_time
            if len(transpiled_program) % minimum_clock_time != 0
            else 0
        )

        phix_q0_waveform = (
            "flux_line_phix_q0",
            np.array(pad_length * [0] + [anneal_step["phix_q0"] for anneal_step in transpiled_program]),
        )
        phiz_q0_waveform = (
            "flux_line_phiz_q0",
            np.array(pad_length * [0] + [anneal_step["phiz_q0"] for anneal_step in transpiled_program]),
        )
        phix_q1_waveform = (
            "flux_line_phix_q1",
            np.array(pad_length * [0] + [anneal_step["phix_q1"] for anneal_step in transpiled_program]),
        )
        phiz_q1_waveform = (
            "flux_line_phiz_q1",
            np.array(pad_length * [0] + [anneal_step["phiz_q1"] for anneal_step in transpiled_program]),
        )
        phix_c0_1_waveform = (
            "flux_line_phix_c0_1",
            np.array(pad_length * [0] + [anneal_step["phix_c0_1"] for anneal_step in transpiled_program]),
        )
        phiz_c0_1_waveform = (
            "flux_line_phiz_c0_1",
            np.array(pad_length * [0] + [anneal_step["phiz_c0_1"] for anneal_step in transpiled_program]),
        )
        assert np.allclose(anneal_waveforms[phix_q0_waveform[0]].envelope(), phix_q0_waveform[1])
        assert np.allclose(anneal_waveforms[phiz_q0_waveform[0]].envelope(), phiz_q0_waveform[1])
        assert np.allclose(anneal_waveforms[phix_q1_waveform[0]].envelope(), phix_q1_waveform[1])
        assert np.allclose(anneal_waveforms[phiz_q1_waveform[0]].envelope(), phiz_q1_waveform[1])
        assert np.allclose(anneal_waveforms[phix_c0_1_waveform[0]].envelope(), phix_c0_1_waveform[1])
        assert np.allclose(anneal_waveforms[phiz_c0_1_waveform[0]].envelope(), phiz_c0_1_waveform[1])

    def test_get_waveforms_same_bus_error(self, annealing_program_transpiled):
        """Test get waveforms method works as intended"""

        # point two fluxes to the same bus
        annealing_program_transpiled._flux_to_bus_topology[0].bus = annealing_program_transpiled._flux_to_bus_topology[
            1
        ].bus
        error_string = f"More than one flux pointing at bus {annealing_program_transpiled._flux_to_bus_topology[1].bus} in the runcard flux to bus topology"
        with pytest.raises(ValueError, match=error_string):
            _ = annealing_program_transpiled.get_waveforms()

    def test_get_waveforms_xtalk(self, annealing_program_transpiled):
        """Test get waveforms method works as intended"""
        crosstalk_matrix = MagicMock()
        with patch.object(FluxVector, "set_crosstalk") as mock_set_crosstalk:
            _ = annealing_program_transpiled.get_waveforms(crosstalk_matrix=crosstalk_matrix)
        # check that __matmul__ is called at each anneal step
        mock_set_crosstalk.assert_called()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import os
import sys

from sphinxawesome_theme.postprocess import Icons

sys.path.insert(0, os.path.abspath(".."))

project = "Qililab"
copyright = "2023, Qilimanjaro"
author = "Qilimanjaro"

import qililab  # noqa: E402, ICN001

release = qililab.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinx_mdinclude",  # allows the mdinclude directive to add Markdown files
    "sphinx.ext.napoleon",  # converts Google docstrings into rst
    "sphinx_automodapi.automodapi",
    # "sphinxawesome_theme.docsearch",  # TODO: Uncomment this when access to DocSearch!
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "releases/*"]
source_suffix = [".rst", ".md"]
pygments_style = "default"

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_preprocess_types = True

automodapi_toctreedirnm = "code/api"  # location where the automodapi rst files are built

autoclass_content = "class"  # only show class docstrings (hide init docstrings)

autodoc_typehints = "none"
autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_title = project
html_permalinks_icon = Icons.permalinks_icon
html_favicon = "_static/q_light.jpeg"
html_baseurl = "https://qililab.com/"
html_theme = "sphinxawesome_theme"
html_sidebars = {"**": ["sidebar_main_nav_links.html", "sidebar_toc.html"]}
html_theme_options = {
    "logo_light": "_static/q_light.jpeg",
    "logo_dark": "_static/q_dark.jpeg",
    "show_prev_next": True,
    "awesome_external_links": True,
    "main_nav_links": {
        "QaaS": "https://qaas.readthedocs.io/en/latest/",
        "Qiboconnection": "https://qaas.readthedocs.io/projects/qiboconnection/en/latest/",
        "Qililab": "index",
    },
    "extra_header_link_icons": {
        "repository on GitHub": {
            "link": "https://github.com/qilimanjaro-tech/qililab",
            "icon": (
                '<svg height="26px" style="margin-top:-2px;display:inline" '
                'viewBox="0 0 45 44" '
                'fill="currentColor" xmlns="http://www.w3.org/2000/svg">'
                '<path fill-rule="evenodd" clip-rule="evenodd" '
                'd="M22.477.927C10.485.927.76 10.65.76 22.647c0 9.596 6.223 17.736 '
                "14.853 20.608 1.087.2 1.483-.47 1.483-1.047 "
                "0-.516-.019-1.881-.03-3.693-6.04 "
                "1.312-7.315-2.912-7.315-2.912-.988-2.51-2.412-3.178-2.412-3.178-1.972-1.346.149-1.32.149-1.32 "
                "2.18.154 3.327 2.24 3.327 2.24 1.937 3.318 5.084 2.36 6.321 "
                "1.803.197-1.403.759-2.36 "
                "1.379-2.903-4.823-.548-9.894-2.412-9.894-10.734 "
                "0-2.37.847-4.31 2.236-5.828-.224-.55-.969-2.759.214-5.748 0 0 "
                "1.822-.584 5.972 2.226 "
                "1.732-.482 3.59-.722 5.437-.732 1.845.01 3.703.25 5.437.732 "
                "4.147-2.81 5.967-2.226 "
                "5.967-2.226 1.185 2.99.44 5.198.217 5.748 1.392 1.517 2.232 3.457 "
                "2.232 5.828 0 "
                "8.344-5.078 10.18-9.916 10.717.779.67 1.474 1.996 1.474 4.021 0 "
                "2.904-.027 5.247-.027 "
                "5.96 0 .58.392 1.256 1.493 1.044C37.981 40.375 44.2 32.24 44.2 "
                '22.647c0-11.996-9.726-21.72-21.722-21.72" '
                'fill="currentColor"/></svg>'
            ),
        },
    },
}
html_static_path = ["_static"]
html_css_files = ["custom.css"]



import os
import tempfile
from unittest.mock import Mock, call, create_autospec

import numpy as np
import pytest

from qililab.platform.platform import Platform
from qililab.qprogram.blocks import ForLoop, Loop
from qililab.qprogram.calibration import Calibration
from qililab.qprogram.crosstalk_matrix import CrosstalkMatrix
from qililab.qprogram.experiment import Experiment
from qililab.qprogram.experiment_executor import ExperimentExecutor
from qililab.qprogram.qprogram import Domain, QProgram
from qililab.result.qprogram import QProgramResults, QuantumMachinesMeasurementResult
from qililab.typings.enums import Parameter
from qililab.waveforms import IQPair, Square


@pytest.fixture(name="platform")
def mock_platform():
    """Fixture to create a mock Platform."""
    qprogram_results = QProgramResults()
    qprogram_results.append_result(
        "readout_bus", QuantumMachinesMeasurementResult(bus="readout", I=np.arange(0, 11), Q=np.arange(100, 111))
    )
    qprogram_results.append_result(
        "readout_bus", QuantumMachinesMeasurementResult(bus="readout", I=np.arange(0, 11), Q=np.arange(100, 111))
    )

    platform = create_autospec(Platform)
    platform.set_parameter = Mock()
    platform.get_parameter = Mock(return_value=1.23)
    platform.execute_qprogram = Mock(return_value=qprogram_results)
    platform.to_dict = Mock(return_value={"name": "platform"})
    platform.experiment_results_base_path = tempfile.gettempdir()
    platform.experiment_results_path_format = "{date}/{time}/{label}.h5"
    platform.set_crosstalk = Mock()

    return platform


@pytest.fixture(name="qprogram")
def fixture_qprogram():
    """Fixture to create a mock QProgram."""
    qp = QProgram()
    gain = qp.variable(label="gain", domain=Domain.Voltage)
    with qp.for_loop(gain, 0, 1.0, 0.1):
        qp.set_gain(bus="readout_bus", gain=gain)
        qp.measure(
            "readout_bus",
            waveform=IQPair(Square(1.0, 40), Square(1.0, 40)),
            weights=IQPair(Square(1.0, 100), Square(1.0, 100)),
        )
    with qp.average(shots=1000):
        with qp.loop(gain, values=np.linspace(0.0, 1.0, 11)):
            qp.set_gain(bus="readout_bus", gain=gain)
            qp.measure(
                "readout_bus",
                waveform=IQPair(Square(1.0, 40), Square(1.0, 40)),
                weights=IQPair(Square(1.0, 100), Square(1.0, 100)),
            )

    return qp


@pytest.fixture(name="crosstalk")
def fixture_crosstalk():
    return CrosstalkMatrix.from_array(
        buses=["flux_q0", "flux_q1"], matrix_array=np.array([[1.47046905, 0.12276261], [-0.55322207, 1.58247856]])
    )


def get_qprogram_nshots_by_loop(nshots: int, qprogram: QProgram):
    with qprogram.average(nshots):
        pass
    return qprogram


def get_qprogram_gain_by_get_parameter(gain: float, qprogram: QProgram):
    qprogram.set_gain(bus="readout_bus", gain=gain)
    return qprogram


def get_qprogram_frequency_and_bias_by_loop(frequency: float, bias: float, qprogram: QProgram):
    qprogram.set_frequency(bus="readout", frequency=frequency)
    qprogram.set_gain(bus="readout", gain=bias)
    return qprogram


@pytest.fixture(name="experiment")
def fixture_experiment(qprogram: QProgram, crosstalk: CrosstalkMatrix):
    """Fixture to create a mock Experiment."""
    experiment = Experiment(label="experiment")
    bias = experiment.variable(label="Bias (mV)", domain=Domain.Voltage)
    frequency = experiment.variable(label="Frequency (Hz)", domain=Domain.Frequency)
    nshots = experiment.variable(label="nshots", domain=Domain.Scalar, type=int)

    # Test SetCrosstalk
    experiment.set_crosstalk(crosstalk=crosstalk)

    # Test SetParameter
    experiment.set_parameter(alias="drive_q0", parameter=Parameter.VOLTAGE, value=0.0)
    experiment.set_parameter(alias="drive_q1", parameter=Parameter.VOLTAGE, value=0.5)
    experiment.set_parameter(alias="drive_q2", parameter=Parameter.VOLTAGE, value=1.0)
    experiment.set_parameter(alias="flux_q0", parameter=Parameter.FLUX, value=0.0)

    # Test GetParameter returns a variable that can be reused
    gain = experiment.get_parameter(alias="flux_q0", parameter=Parameter.GAIN)
    # Reused in SetParameter
    experiment.set_parameter(alias="flux_q1", parameter=Parameter.GAIN, value=gain)
    # Reused in ExecuteQProgram
    experiment.execute_qprogram(lambda gain=gain: get_qprogram_gain_by_get_parameter(gain=gain, qprogram=qprogram))

    # Test inner loops
    with experiment.for_loop(bias, 0.0, 1.0, 0.5):
        experiment.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=bias)
        with experiment.loop(frequency, values=np.array([2e9, 3e9])):
            experiment.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=frequency)
            experiment.execute_qprogram(qprogram)
    # Test parallel loops
    with experiment.parallel(loops=[ForLoop(bias, 0.0, 1.0, 0.5), Loop(frequency, values=np.array([2e9, 3e9, 4e9]))]):
        experiment.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=bias)
        experiment.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=frequency)
        experiment.execute_qprogram(qprogram)

    # Test qprogram lambda with variable from loop
    with experiment.for_loop(nshots, 0, 3):
        experiment.execute_qprogram(lambda nshots=nshots: get_qprogram_nshots_by_loop(nshots=nshots, qprogram=qprogram))  # type: ignore

    # Test qprogram lambda with two variables from loop
    with experiment.parallel(loops=[ForLoop(bias, 0.0, 1.0, 0.5), Loop(frequency, values=np.array([2e9, 3e9, 4e9]))]):
        experiment.execute_qprogram(
            lambda frequency=frequency, bias=bias: get_qprogram_frequency_and_bias_by_loop(
                frequency=frequency, bias=bias, qprogram=qprogram
            )
        )

    return experiment


class TestExperimentExecutor:
    """Test ExperimentExecutor class"""

    def test_execute(self, platform, experiment, qprogram, crosstalk):
        """Test the execute method to ensure the experiment is executed correctly and results are stored."""
        executor = ExperimentExecutor(platform=platform, experiment=experiment)
        resuls_path = executor.execute()

        # Check if the correct file path is returned
        assert resuls_path.startswith(os.path.abspath(tempfile.gettempdir()))
        assert resuls_path.endswith(".h5")

        # Check that platform methods were called in the correct order
        expected_calls = [
            call.to_dict(),
            call.set_crosstalk(crosstalk=crosstalk),
            call.set_parameter(alias="drive_q0", parameter=Parameter.VOLTAGE, value=0.0, channel_id=None),
            call.set_parameter(alias="drive_q1", parameter=Parameter.VOLTAGE, value=0.5, channel_id=None),
            call.set_parameter(alias="drive_q2", parameter=Parameter.VOLTAGE, value=1.0, channel_id=None),
            call.set_parameter(alias="flux_q0", parameter=Parameter.FLUX, value=0.0, channel_id=None),
            # Check that get_parameter returns a variable that can be reused
            call.get_parameter(alias="flux_q0", parameter=Parameter.GAIN, channel_id=None),
            call.set_parameter(alias="flux_q1", parameter=Parameter.GAIN, value=1.23, channel_id=None),
            call.execute_qprogram(
                qprogram=get_qprogram_gain_by_get_parameter(gain=1.23, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            # Start of nested loops
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=0.0, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=2e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=3e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=0.5, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=2e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=3e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=1.0, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=2e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=3e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            # End of nested loops
            # Start of parallel loop
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=0.0, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=2e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=0.5, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=3e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            call.set_parameter(alias="readout_bus", parameter=Parameter.VOLTAGE, value=1.0, channel_id=None),
            call.set_parameter(alias="readout_bus", parameter=Parameter.LO_FREQUENCY, value=4e9, channel_id=None),
            call.execute_qprogram(qprogram=qprogram, bus_mapping=None, calibration=None, debug=False),
            # End of parallel loop
            # Start of nshots loop with lambda execution
            call.execute_qprogram(
                qprogram=get_qprogram_nshots_by_loop(nshots=0, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            call.execute_qprogram(
                qprogram=get_qprogram_nshots_by_loop(nshots=1, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            call.execute_qprogram(
                qprogram=get_qprogram_nshots_by_loop(nshots=2, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            # End of nshots loop
            # Start of parallel loop with lambda execution
            call.execute_qprogram(
                qprogram=get_qprogram_frequency_and_bias_by_loop(frequency=2e9, bias=0.0, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            call.execute_qprogram(
                qprogram=get_qprogram_frequency_and_bias_by_loop(frequency=3e9, bias=0.5, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            call.execute_qprogram(
                qprogram=get_qprogram_frequency_and_bias_by_loop(frequency=4e9, bias=1.0, qprogram=qprogram),
                bus_mapping=None,
                calibration=None,
                debug=False,
            ),
            # End of parallel loop
        ]

        # If you want to ensure the exact sequence across all calls
        platform.assert_has_calls(expected_calls, any_order=False)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Result class."""
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from qililab.result.counts import Counts
from qililab.typings.enums import ResultName
from qililab.typings.factory_element import FactoryElement
from qililab.utils import nested_dict_to_pandas_dataframe


class Result(FactoryElement, ABC):
    """Class used to hold the results of a single execution."""

    name: ResultName
    """Name of the result."""
    data_dataframe_indices: set[str]
    """Set of indices that are used in the data DataFrame."""

    def counts(self) -> dict:
        """(Only for `Qblox`) Returns a Counts object containing the counts of each state.

        Returns:
            Counts: Counts object containing the counts of each state.

        Raises:
            NotImplementedError: this method is not implemented for n measurements on the same qubit
        """
        raise NotImplementedError

    def samples(self) -> np.ndarray:
        """(Only for `Qblox`) Returns an array containing the measured samples.

        The shape of the returned array is ``(# sequencers, # bins)``.

        Returns:
            np.ndarray: An array containing the measured samples (0 or 1).

        Raises:
            NotImplementedError: this method is not implemented for n measurements on the same qubit
        """
        raise NotImplementedError

    def acquisitions(self) -> pd.DataFrame:
        """Return acquisition values.

        Returns:
            pd.DataFrame: I, Q, amplitude and phase.
        """
        raise NotImplementedError

    def probabilities(self) -> dict[str, float]:
        """(Only for `Qblox`) Return probabilities of being in the ground and excited state.

        Returns:
            dict[str, float]: Dictionary containing the quantum states as the keys of the dictionary, and the
                probabilities obtained for each state as the values of the dictionary.
        """
        return self.counts_object().probabilities()

    def counts_object(self) -> Counts:
        """(Only for `Qblox`) Returns a Counts object containing the amount of times each state was measured.

        Raises:
            NotImplementedError: Not implemented.

        Returns:
            Counts: Counts object containing the amount of times each state was measured.
        """
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict:
        """
        Returns:
            dict: Dictionary containing all the class information.
        """

    def to_dataframe(self) -> pd.DataFrame:
        """
        Returns:
            DataFrame: dataframe containing all the results.
        """

        return nested_dict_to_pandas_dataframe(self.to_dict())

    @property
    @abstractmethod
    def array(self) -> np.ndarray:
        """Returns the results in a numpy array format.

        Returns:
            np.ndarray: Numpy array containing the results.
        """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instruments module"""

from .cluster import Cluster
from .device import Device
from .keithley_2600 import Keithley2600Driver
from .mini_circuits import MiniCircuitsDriver
from .qblox_d5a import QbloxD5a
from .qblox_s4g import QbloxS4g
from .qcm_qrm import QcmQrm
from .qdevil_qdac2 import QDevilQDac2
from .qmm_driver import QMMDriver
from .rohde_schwarz import RohdeSchwarzSGS100A
from .yokogawa_gs200 import YokogawaGS200

__all__ = [
    "Cluster",
    "Device",
    "Keithley2600Driver",
    "MiniCircuitsDriver",
    "QDevilQDac2",
    "QMMDriver",
    "QbloxD5a",
    "QbloxS4g",
    "QcmQrm",
    "RohdeSchwarzSGS100A",
    "YokogawaGS200",
]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for usual dataframe manipulation"""

import pandas as pd


def concatenate_creating_new_name_index(dataframes: list[pd.DataFrame], new_index_name: str) -> pd.DataFrame:
    """Concatenates an ordered list of dataframes into a single one, adding a new named column containing the index of
     the dataframe the data came from in the original list.
     The result dataframe will have as columns the union of the original dataframes plus the new column with the
     indices.
     The index structure of the dataframes will not be considered when creating the new one (as opposed to when
     normally using a dataframe, where you would get a multiindex if previous index was structured).

     Args:
         dataframes: list of dataframes to concatenate
         new_index_name: name to be given to the new column containing the indices

    Returns:
        pd.Dataframe: the new dataframe
    """
    concatenated_df = pd.concat(dataframes, ignore_index=True)
    concatenated_df.index.rename(new_index_name, inplace=True)
    concatenated_df.reset_index(inplace=True)
    return concatenated_df



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CircuitToPulses class"""

from dataclasses import asdict

import numpy as np
from qibo import gates

from qililab.constants import RUNCARD
from qililab.digital.native_gates import _GateHandler
from qililab.pulse.pulse import Pulse
from qililab.pulse.pulse_event import PulseEvent
from qililab.pulse.pulse_schedule import PulseSchedule
from qililab.settings.digital.digital_compilation_settings import DigitalCompilationSettings
from qililab.settings.digital.gate_event_settings import GateEventSettings
from qililab.typings.enums import Line
from qililab.utils import Factory

from .native_gates import Drag, Wait


class CircuitToPulses:
    """Translates circuits into pulse sequences.

    Args:
        settings (DigitalCompilationSettings): Object containing the Digital Compilations Settings and the info on chip's physical qubits.
            It can be obtained from the ``digital_compilation_settings`` attribute of a ``Platform`` object.
    """

    def __init__(self, settings: DigitalCompilationSettings):
        self.settings: DigitalCompilationSettings = settings
        """Object containing the digital compilations settings and the info on chip's physical qubits."""

    def run(self, circuit_gates: list[gates.Gate]) -> PulseSchedule:
        """Translates a Qibo circuit into its corresponding pulse sequences.

        Check public docstring in :meth:`.CircuitTranspiler.gates_to_pulses()` for more information.

        Args:
            circuit_gates (list[gates.Gate]): list of native gates of the qibo circuit.

        Returns:
            PulseSequences: equivalent :class:`PulseSequences` class.
        """

        pulse_schedule: PulseSchedule = PulseSchedule()
        time: dict[int, int] = {}  # init/restart time
        for gate in circuit_gates:
            # handle wait gates
            if isinstance(gate, Wait):
                self._update_time(time=time, qubit=gate.qubits[0], gate_time=gate.parameters[0])
                continue

            # Measurement gates need to be handled on their own because qibo allows to define
            # an M gate as eg. gates.M(*range(5))
            if isinstance(gate, gates.M):
                gate_schedule = []
                gate_qubits = gate.qubits
                for qubit in gate_qubits:
                    gate_schedule += self._gate_schedule_from_settings(gates.M(qubit))

            # handle control gates
            else:
                # extract gate schedule
                gate_schedule = self._gate_schedule_from_settings(gate)
                gate_qubits = self._get_gate_qubits(gate, gate_schedule)

            # process gate_schedule to pulses for both M and control gates
            # get total duration for the gate
            gate_time = self._get_total_schedule_duration(gate_schedule)
            # update time, start time is that of the qubit most ahead in time
            start_time = 0
            for qubit in gate_qubits:
                start_time = max(self._update_time(time=time, qubit=qubit, gate_time=gate_time), start_time)
            # sync gate end time
            self._sync_qubit_times(gate_qubits, time=time)
            # apply gate schedule
            for gate_event in gate_schedule:
                # add control gate schedule
                pulse_event = self._gate_element_to_pulse_event(time=start_time, gate=gate, gate_event=gate_event)
                # pop first qubit from gate if it is measurement
                # this is so that the target qubit for multiM gates is every qubit in the M gate
                if isinstance(gate, gates.M):
                    gate = gates.M(*gate.qubits[1:])
                # add event
                delay = self.settings.buses[gate_event.bus].delay
                pulse_schedule.add_event(pulse_event=pulse_event, bus_alias=gate_event.bus, delay=delay)  # type: ignore

        for bus_alias in self.settings.buses:
            # If we find a flux port, create empty schedule for that port.
            # This is needed because for Qblox instrument working in flux buses as DC sources, if we don't
            # add an empty schedule its offsets won't be activated and the results will be misleading.
            if self.settings.buses[bus_alias].line == Line.FLUX:
                pulse_schedule.create_schedule(bus_alias=bus_alias)

        return pulse_schedule

    def _gate_schedule_from_settings(self, gate: gates.Gate) -> list[GateEventSettings]:
        """Gets the gate schedule. The gate schedule is the list of pulses to apply
        to a given bus for a given gate

        Args:
            gate (Gate): Qibo gate

        Returns:
            list[GateEventSettings]: schedule list with each of the pulses settings
        """

        gate_schedule = self.settings.get_gate(name=gate.__class__.__name__, qubits=gate.qubits)

        if not isinstance(gate, Drag):
            return gate_schedule

        # drag gates are currently the only parametric gates we are handling and they are handled here
        if len(gate_schedule) > 1:
            raise ValueError(
                f"Schedule for the drag gate is expected to have only 1 pulse but instead found {len(gate_schedule)} pulses"
            )
        drag_schedule = GateEventSettings(
            **asdict(gate_schedule[0])
        )  # make new object so that gate_schedule is not overwritten
        theta = _GateHandler.normalize_angle(angle=gate.parameters[0])
        amplitude = drag_schedule.pulse.amplitude * theta / np.pi
        phase = _GateHandler.normalize_angle(angle=gate.parameters[1])
        if amplitude < 0:
            amplitude = -amplitude
            phase = _GateHandler.normalize_angle(angle=gate.parameters[1] + np.pi)
        drag_schedule.pulse.amplitude = amplitude
        drag_schedule.pulse.phase = phase
        return [drag_schedule]

    @staticmethod
    def _get_total_schedule_duration(schedule: list[GateEventSettings]) -> int:
        """Returns total time for a gate schedule. This is done by taking the max of (init + duration)
        for all the elements in the schedule

        Args:
            schedule (list[CircuitPulseSettings]): Schedule of pulses to apply

        Returns:
            int: Total gate time
        """
        time = 0
        for schedule_element in schedule:
            time = max(time, schedule_element.pulse.duration + schedule_element.wait_time)
        return time

    def _get_gate_qubits(self, gate: gates.Gate, schedule: list[GateEventSettings] | None = None) -> tuple[int, ...]:
        """Gets qubits involved in gate. This includes gate.qubits but also qubits which are targets of
        buses in the gate schedule

        Args:
            schedule (list[CircuitPulseSettings]): Gate schedule

        Returns:
            list[int]: list of qubits
        """

        schedule_qubits = (
            [
                qubit
                for schedule_element in schedule
                for qubit in self.settings.buses[schedule_element.bus].qubits
                if schedule_element.bus in self.settings.buses
            ]
            if schedule is not None
            else []
        )

        gate_qubits = list(gate.qubits)

        return tuple(set(schedule_qubits + gate_qubits))  # convert to set and back to list to remove repeated items

    def _gate_element_to_pulse_event(self, time: int, gate: gates.Gate, gate_event: GateEventSettings) -> PulseEvent:
        """Translates a gate element into a pulse.

        Args:
            time (dict[int, int]): dictionary containing qubit indices as keys and current time (ns) as values
            gate (gate): circuit gate. This is used only to know the qubit target of measurement gates
            gate_event (GateEventSettings): gate event, a single element of a gate schedule containing information
            about the pulse to be applied
            bus (bus): bus through which the pulse is sent

        Returns:
            PulseEvent: pulse event corresponding to the input gate event
        """

        # copy to avoid modifying runcard settings
        pulse = gate_event.pulse
        pulse_shape_copy = pulse.shape.copy()
        pulse_shape = Factory.get(pulse_shape_copy.pop(RUNCARD.NAME))(**pulse_shape_copy)

        # handle measurement gates and target qubits for control gates which might have multi-qubit schedules
        bus = self.settings.buses[gate_event.bus]
        qubit = (
            gate.qubits[0]
            if isinstance(gate, gates.M)
            else next((qubit for qubit in bus.qubits), None)
            if bus is not None
            else None
        )

        return PulseEvent(
            pulse=Pulse(
                amplitude=pulse.amplitude,
                phase=pulse.phase,
                duration=pulse.duration,
                frequency=0,
                pulse_shape=pulse_shape,
            ),
            start_time=time + gate_event.wait_time + self.settings.delay_before_readout,
            pulse_distortions=bus.distortions,
            qubit=qubit,
        )

    def _update_time(self, time: dict[int, int], qubit: int, gate_time: int):
        """Creates new timeline if not already created and update time.

        Args:
            time (Dict[int, int]): Dictionary with the time of each qubit.
            qubit_idx (int): qubit index
            gate_time (int): total duration of the gate
        """
        if qubit not in time:
            time[qubit] = 0
        old_time = time[qubit]
        residue = (gate_time) % self.settings.minimum_clock_time
        if residue != 0:
            gate_time += self.settings.minimum_clock_time - residue
        time[qubit] += gate_time
        return old_time

    @staticmethod
    def _sync_qubit_times(qubits: list[int], time: dict[int, int]):
        """Syncs the time of the given qubit list

        Args:
            qubits (list[int]): qubits to sync
            time (dict[int, int]): time dictionary
        """
        max_time = max((time[qubit] for qubit in qubits if qubit in time), default=0)
        for qubit in qubits:
            time[qubit] = max_time



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Bins Acquisitions Result"""

from dataclasses import dataclass

import numpy as np

from qililab.result.acquisitions import Acquisitions
from qililab.result.counts import Counts
from qililab.result.qblox_results.bins_data import BinsData
from qililab.result.qblox_results.qblox_bins_acquisition import QbloxBinAcquisition


@dataclass
class QbloxBinsAcquisitions(Acquisitions):
    """Qblox Bins Acquisitions Results
    Args:
        bins (list[BinsData]): List containing a BinsData for each sequencer.
        integration_lengths (list[int]): List of integration lengths for each sequencer.
    """

    bins: list[BinsData]
    integration_lengths: list[int]

    def __post_init__(self):
        """Create acquisitions"""
        self._acquisitions = [
            self._build_bin_acquisition(bins_data=bins_data, integration_length=self.integration_lengths[sequencer_id])
            for sequencer_id, bins_data in enumerate(self.bins)
        ]
        self.data_dataframe_indices = set().union(*[acq.data_dataframe_indices for acq in self._acquisitions])

    def _build_bin_acquisition(self, bins_data: BinsData, integration_length: int):
        """build a bin acquisition"""
        i_values = np.array(bins_data.integration.path0, dtype=np.float32)
        q_values = np.array(bins_data.integration.path1, dtype=np.float32)
        return QbloxBinAcquisition(integration_length=integration_length, i_values=i_values, q_values=q_values)

    def counts(self) -> Counts:
        """Return the counts of measurements in each state.

        Returns:
            Counts: Counts object with the number of measurements in that state.
        """
        # Check that all sequencers have the same number of bins.
        if any(len(seq_bins) != (num_bins := len(self.bins[0])) for seq_bins in self.bins):
            raise IndexError("Sequencers must have the same number of bins to compute the counts.")
        # TODO: Add limitations to check we are doing single-shot for multi qubit?
        counts_object = Counts(n_qubits=len(self.bins))
        for seq in self.bins:
            if any(bin_avg > 1 for bin_avg in seq.avg_cnt):
                raise ValueError("Counts cannot be used when using a hardware average higher than 1.")
        for bin_idx in range(num_bins):
            # The threshold inside of a qblox bin is the name they use for already classified data as a value between
            # 0 and 1, not the value used in the comparator to perform such classification.
            measurement_as_list = [int(bins_data.threshold[bin_idx]) for bins_data in self.bins]
            measurement = "".join(str(bit) for bit in measurement_as_list)
            counts_object.add_measurement(state=measurement)
        return counts_object

    def samples(self) -> np.ndarray:
        """Returns an array containing the samples measured in each bin of each sequencer.

        The shape of the returned array is ``(# sequencers, # bins)``.

        Returns:
            np.ndarray: An array containing the samples measured in each bin of each sequencer.
        """
        # Check that all sequencers have the same number of bins.
        if any(len(seq_bins) != len(self.bins[0]) for seq_bins in self.bins):
            raise IndexError("Sequencers must have the same number of bins to return the samples.")
        samples = [seq_data.threshold for seq_data in self.bins]
        return np.array(samples)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Connection class"""
from abc import ABC
from dataclasses import asdict, dataclass
from functools import partial
from typing import Callable

from qililab.config import logger
from qililab.constants import RUNCARD
from qililab.typings import ConnectionName, Device, FactoryElement


class Connection(ABC, FactoryElement):
    """Abstract base class declaring the necessary attributes and methods for a connection.

    Args:
        name (ConnectionName): type of the Connection.
        settings (ConnectionSettings): Settings of the Connection.
        _connected (bool): True if the instrument is connected.
        _device (Device | None): Device driver to connect to the Instrument.
        _device_name (str): Name of the device to connect to.
    """

    @dataclass
    class ConnectionSettings:
        """Contains the settings of a connection.

        Args:
            address (str): Connection address of the connected device.
        """

        address: str

    name: ConnectionName
    settings: ConnectionSettings  # a subtype of settings must be specified by the subclass

    class CheckConnected:
        """Property used to check if the connection has established with an instrument."""

        def __init__(self, method: Callable):
            self._method = method

        def __get__(self, obj, objtype):
            """Support instance methods."""
            return partial(self.__call__, obj)

        def __call__(self, ref: "Connection", *args, **kwargs):
            """
            Args:
                method (Callable): Class method.

            Raises:
                AttributeError: If connection has not been established with an instrument.
            """
            if not ref._connected:
                raise AttributeError("Instrument is not connected")
            if not hasattr(ref, "_device"):
                raise AttributeError("Instrument Device has not been initialized")
            return self._method(ref, *args, **kwargs)

    def __init__(self, settings: dict):
        self.settings = self.ConnectionSettings(**settings)
        self._connected = False
        self._device: Device | None = None
        self._device_name = ""

    def connect(self, device: Device, device_name: str):
        """Establish connection with the instrument. Initialize self._device variable."""
        if self._connected:
            raise ValueError(f"Instrument ({device_name}) is already connected")
        self._connected = True
        self._device = device
        self._device_name = device_name
        logger.info("Connecting to instrument %s.", self._device_name)

    @CheckConnected
    def close(self):
        """Close an open connection with the instrument. Set self._device to None"""
        logger.info("Closing instrument %s.", self._device_name)
        self._device.close()
        self._connected = False
        self._device = None
        self._device_name = ""

    @property
    def address(self):
        """Connection 'address' property.

        Returns:
            str: settings.address.
        """
        return self.settings.address

    @CheckConnected
    def check_instrument_is_connected(self):
        """check if instrument has been connected and initialized.
        It is checked via a decorator.
        """

    def to_dict(self):
        """Return a dict representation of the Connection class."""
        return {RUNCARD.NAME: self.name.value} | asdict(self.settings)



"""Test execute script"""
from unittest.mock import MagicMock, patch

import pytest
from qibo.gates import CNOT, RY, H, M, X
from qibo.models import Circuit

import qililab as ql


class TestExecute:
    """Test execute method in the `execute.py` script"""

    def test_execute(self):
        n_qubits = 5
        circuit = Circuit(n_qubits)
        circuit.add([X(1), H(2), RY(0, 2), CNOT(4, 1), X(4), H(3), M(*range(5))])
        runcard = "galadriel.yml"
        mock_platform = MagicMock()
        with patch("qililab.execute_circuit.build_platform", return_value=mock_platform) as mock_build:
            results = ql.execute(circuit, runcard)
            assert isinstance(results, MagicMock)
            mock_build.assert_called_with(runcard="galadriel.yml")
        mock_platform.connect.assert_called_once_with()
        mock_platform.initial_setup.assert_called_once_with()
        mock_platform.turn_on_instruments.assert_called_once_with()
        mock_platform.execute.assert_called_once()
        mock_platform.disconnect.assert_called_once_with()

    def test_execute_calls_disconnect_after_error(self):
        n_qubits = 5
        circuit = Circuit(n_qubits)
        circuit.add([X(1), H(2), RY(0, 2), CNOT(4, 1), X(4), H(3), M(*range(5))])
        runcard = "galadriel.yml"
        with pytest.raises(Exception):
            mock_platform = MagicMock()
            mock_platform.turn_on_instruments.side_effect = Exception()
            with patch("qililab.execute_circuit.build_platform", return_value=mock_platform) as mock_build:
                ql.execute(circuit, runcard)
            mock_build.assert_called_with(runcard="galadriel.yml")
            mock_platform.connect.assert_called_once_with()
            mock_platform.initial_setup.assert_called_once_with()
            mock_platform.turn_on_instruments.assert_called_once_with()
            mock_platform.execute.assert_not_called()
            mock_platform.disconnect.assert_called_once_with()

    def test_execute_circuit_with_multiple_circuits(self):
        """Test that executing a list of circuits returns a list of results."""
        n_qubits = 5
        circuit = Circuit(n_qubits)
        circuit.add([X(1), H(2), RY(0, 2), CNOT(4, 1), X(4), H(3), M(*range(5))])
        runcard = "galadriel.yml"
        mock_platform = MagicMock()
        with patch("qililab.execute_circuit.build_platform", return_value=mock_platform) as mock_build:
            results = ql.execute([circuit, circuit], runcard)
            assert isinstance(results, list)
            mock_build.assert_called_with(runcard="galadriel.yml")
        mock_platform.connect.assert_called_once_with()
        mock_platform.initial_setup.assert_called_once_with()
        mock_platform.turn_on_instruments.assert_called_once_with()
        assert mock_platform.execute.call_count == 2
        mock_platform.disconnect.assert_called_once_with()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import re
from dataclasses import asdict, dataclass

from qililab.constants import GATE_ALIAS_REGEX
from qililab.settings.digital.digital_compilation_bus_settings import DigitalCompilationBusSettings
from qililab.settings.digital.gate_event_settings import GateEventSettings
from qililab.typings import ChannelID, Parameter, ParameterValue
from qililab.utils.asdict_factory import dict_factory


@dataclass
class DigitalCompilationSettings:
    """Dataclass with all the settings and gates definitions needed to decompose gates into pulses."""

    minimum_clock_time: int
    delay_before_readout: int
    topology: list[tuple[int, int]]
    gates: dict[str, list[GateEventSettings]]
    buses: dict[str, DigitalCompilationBusSettings]

    def __post_init__(self):
        """Build the Gates Settings based on the master settings."""
        self.topology = [tuple(element) if isinstance(element, list) else element for element in self.topology]
        self.gates = {gate: [GateEventSettings(**event) for event in schedule] for gate, schedule in self.gates.items()}
        self.buses = {bus: DigitalCompilationBusSettings(**settings) for bus, settings in self.buses.items()}

    def to_dict(self):
        """Serializes gate settings to dictionary and removes fields with None values"""

        return asdict(self, dict_factory=dict_factory) | {
            "buses": {bus: bus_settings.to_dict() for bus, bus_settings in self.buses.items()}
        }

    def get_gate(self, name: str, qubits: int | tuple[int, int] | tuple[int]):
        """Get gates settings from runcard for a given gate name and qubits.

        Args:
            name (str): Name of the gate.
            qubits (int |  tuple[int, int] | tuple[int]): The qubits the gate is acting on.

        Raises:
            ValueError: If no gate is found.

        Returns:
            GatesSettings: gate settings.
        """

        gate_qubits = (
            (qubits,) if isinstance(qubits, int) else qubits
        )  # tuplify so that the join method below is general
        gate_name = f"{name}({', '.join(map(str, gate_qubits))})"
        gate_name_t = f"{name}({', '.join(map(str, gate_qubits[::-1]))})"

        # parse spaces in tuple if needed, check first case with spaces since it is more common
        if gate_name.replace(" ", "") in self.gates.keys():
            return self.gates[gate_name.replace(" ", "")]
        if gate_name in self.gates.keys():
            return self.gates[gate_name]
        if gate_name_t.replace(" ", "") in self.gates.keys():
            return self.gates[gate_name_t.replace(" ", "")]
        if gate_name_t in self.gates.keys():
            return self.gates[gate_name_t]
        raise KeyError(f"Gate {name} for qubits {qubits} not found in settings.")

    @property
    def gate_names(self) -> list[str]:
        """GatesSettings 'gate_names' property.

        Returns:
            list[str]: List of the names of all the defined gates.
        """
        return list(self.gates.keys())

    def set_parameter(
        self, alias: str, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None
    ):
        """Cast the new value to its corresponding type and set the new attribute.

        Args:
            parameter (Parameter): Name of the parameter to get.
            value (float | str | bool): New value to set in the parameter.
            channel_id (int, optional): Channel id. Defaults to None.
            alias (str): String which specifies where the parameter can be found.
        """
        if parameter == Parameter.DELAY_BEFORE_READOUT:
            self.delay_before_readout = int(value)
            return
        if parameter == Parameter.DELAY:
            if alias not in self.buses:
                raise ValueError(f"Could not find bus {alias} in gate settings.")
            self.buses[alias].delay = int(value)
            return
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if regex_match is None:
            raise ValueError(f"Alias {alias} has incorrect format")
        name = regex_match["gate"]
        qubits_str = regex_match["qubits"]
        qubits = ast.literal_eval(qubits_str)
        gates_settings = self.get_gate(name=name, qubits=qubits)
        schedule_element = 0 if len(alias.split("_")) == 1 else int(alias.split("_")[1])
        gates_settings[schedule_element].set_parameter(parameter, value)

    def get_parameter(self, alias: str, parameter: Parameter, channel_id: int | str | None = None):
        """Get parameter from gate settings.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int, optional): Channel id. Defaults to None.
            alias (str): String which specifies where the parameter can be found.
        """
        # if alias is None or alias == "platform":
        #     return super().get_parameter(parameter=parameter, channel_id=channel_id)
        if parameter == Parameter.DELAY_BEFORE_READOUT:
            return self.delay_before_readout
        if parameter == Parameter.DELAY:
            if alias not in self.buses:
                raise ValueError(f"Could not find bus {alias} in gate settings.")
            return self.buses[alias].delay
        regex_match = re.search(GATE_ALIAS_REGEX, alias)
        if regex_match is None:
            raise ValueError(f"Could not find gate {alias} in gate settings.")
        name = regex_match["gate"]
        qubits_str = regex_match["qubits"]
        qubits = ast.literal_eval(qubits_str)
        gates_settings = self.get_gate(name=name, qubits=qubits)
        schedule_element = 0 if len(alias.split("_")) == 1 else int(alias.split("_")[1])
        return gates_settings[schedule_element].get_parameter(parameter)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Qblox Bin Acquisition Result """

from dataclasses import dataclass

import pandas as pd

from qililab.constants import RESULTSDATAFRAME
from qililab.result.acquisition import Acquisition


@dataclass
class QbloxScopeAcquisition(Acquisition):
    """Qblox Bin Acquisition normalized"""

    def _create_acquisition(self) -> pd.DataFrame:
        """transposes each of the acquired results arrays so that we have for each value
        a structure with i, q, amplitude, phase.
        """
        acquisition_dataframe = super()._create_acquisition()
        acquisition_dataframe.index.rename(RESULTSDATAFRAME.SCOPE_INDEX, inplace=True)
        return acquisition_dataframe



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Circuit Transpiler class"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import networkx as nx

from qililab.digital.circuit_optimizer import CircuitOptimizer
from qililab.digital.circuit_router import CircuitRouter
from qililab.digital.circuit_to_pulses import CircuitToPulses

from .gate_decompositions import translate_gates

if TYPE_CHECKING:
    from qibo import Circuit, gates
    from qibo.transpiler.placer import Placer
    from qibo.transpiler.router import Router

    from qililab.pulse.pulse_schedule import PulseSchedule
    from qililab.settings.digital.digital_compilation_settings import DigitalCompilationSettings


@dataclass
class DigitalTranspilationConfig:
    """Dataclass containing the digital transpilation configuration. Used in the :meth:`.CircuitTranspiler.transpile_circuit()` method"""

    routing: bool = False  # TODO: Change to True, when user confirms it works well.
    """(bool, optional): Whether to route the circuit. Defaults to False."""

    placer: Placer | type[Placer] | tuple[type[Placer], dict] | None = None
    """(Placer | type[Placer] | tuple[type[Placer], dict], optional): ``Placer`` instance, or subclass ``type[Placer]`` to
        use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``ReverseTraversal``."""

    router: Router | type[Router] | tuple[type[Router], dict] | None = None
    """(Router | type[Router] | tuple[type[Router], dict], optional): ``Router`` instance, or subclass ``type[Router]`` to
        use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``Sabre``."""

    routing_iterations: int = 10
    """(int, optional): Number of times to repeat the routing pipeline, to get the best stochastic result. Defaults to 10."""

    optimize: bool = False  # TODO: Maybe also change to True, when user confirms it works well.
    """(bool, optional): Whether to optimize the circuit and/or transpilation. Defaults to False."""

    @property
    def _attributes_ordered(self) -> tuple:
        """Returns the attributes of the dataclass in order, as a tuple."""
        return self.routing, self.placer, self.router, self.routing_iterations, self.optimize


class CircuitTranspiler:
    """Handles circuit transpilation. It has 3 accessible methods:

    - ``circuit_to_native``: transpiles a qibo circuit to native gates (Drag, CZ, Wait, M) and optionally RZ if optimize=False (optimize=True by default)
    - ``circuit_to_pulses``: transpiles a native gate circuit to a ``PulseSchedule``
    - ``transpile_circuit``: runs both of the methods above sequentially

    Args:
        settings (DigitalCompilationSettings): Object containing the Digital Compilations Settings and the info on chip's physical qubits.
            It can be obtained from the :attr:`.Platform.digital_compilation_settings` attribute.
    """

    def __init__(self, settings: DigitalCompilationSettings):
        self.settings: DigitalCompilationSettings = settings
        """Object containing the digital compilations settings and the info on chip's physical qubits."""

        self.optimizer: CircuitOptimizer = CircuitOptimizer(self.settings)
        """Object to do do the complex circuit manipulations with."""

    def transpile_circuit(
        self,
        circuit: Circuit,
        transpilation_config: DigitalTranspilationConfig | None = None,
    ) -> tuple[PulseSchedule, list[int] | None]:
        """Transpiles a list of ``qibo.models.Circuit`` objects into a list of pulse schedules.

        The process involves the following steps (by default only: **3.**, **4**., and **6.** run):


        1. **Routing and Placement:** Routes and places the circuit's logical qubits onto the chip's physical qubits. The final qubit layout is returned and logged. This step uses the ``placer``, ``router``, and ``routing_iterations`` parameters if provided; otherwise, default values are applied. Refer to the :meth:`.CircuitTranspiler.route_circuit()` method for more information.

        2. **1st Optimization:** Canceling adjacent pairs of Hermitian gates (H, X, Y, Z, CNOT, CZ, and SWAPs). Refer to the :meth:`.CircuitTranspiler.optimize_gates()` method for more information.

        3. **Native Gate Translation:** Translates the circuit into the chip's native gate set (CZ, RZ, Drag, Wait, and M (Measurement)). Refer to the :meth:`.CircuitTranspiler.gates_to_native()` method for more information.

        4. **Adding phases to our Drag gates:** commuting RZ gates until the end of the circuit to discard them as virtual Z gates, and due to the phase corrections from CZ. Refer to the :meth:`.CircuitTranspiler.add_phases_from_RZs_and_CZs_to_drags()` method for more information.

        5. **2nd Optimization:** Optimizing the resulting Drag gates, by combining multiple pulses into a single one. Refer to the :meth:`.CircuitTranspiler.optimize_transpiled_gates()` method for more information.

        6. **Pulse Schedule Conversion:** Converts the native gates into a pulse schedule using calibrated settings from the runcard. Refer to the :meth:`.CircuitTranspiler.gates_to_pulses()` method for more information.

        .. note::

            Default steps are only: **3.**, **4**., and **6.**, since they are always needed.

            To do Step **1.** set ``routing=True`` (default behavior skips it).

            To do Steps **2.** and **5.** set ``optimize=True`` (default behavior skips it).

        **Examples:**

        If we instantiate some ``Circuit``, ``Platform`` and ``CircuitTranspiler`` objects like:

        .. code-block:: python

            from qibo import gates
            from qibo.models import Circuit
            from qibo.transpiler.placer import ReverseTraversal, Random
            from qibo.transpiler.router import Sabre
            from qililab import build_platform
            from qililab.digital import CircuitTranspiler

            # Create circuit:
            c = Circuit(5)
            c.add(gates.CNOT(1, 0))

            # Create platform:
            platform = build_platform(runcard="<path_to_runcard>")

            # Create transpiler:
            transpiler = CircuitTranspiler(platform.digital_compilation_settings)

        Now we can transpile like, in the following examples:

        .. code-block:: python

            # Default Transpilation (with ReverseTraversal, Sabre, platform's connectivity and optimize = True):
            transpiled_circuit, final_layouts = transpiler.transpile_circuit(c)

            # Or another case, not doing optimization for some reason, and with Non-Default placer:
            transpiled_circuit, final_layout = transpiler.transpile_circuit(c, placer=Random, optimize=False)

            # Or also specifying the `router` with kwargs:
            transpiled_circuit, final_layouts = transpiler.transpile_circuit(c, router=(Sabre, {"lookahead": 2}))

        .. note::

            Check :ref:`Transpilation <transpilation>`, for more examples of how ``execute()``'s methods automatically apply this.

        Args:
            circuit (Circuit): Qibo circuit.
            transpilation_config (DigitalTranspilationConfig, optional): :class:`.DigitalTranspilationConfig` dataclass containing
                the configuration used during transpilation. Defaults to ``None`` (not changing any default value).
                Check the class:`.DigitalTranspilationConfig` documentation for the keys and values it can contain.

        Returns:
            tuple[PulseSchedule, list[int] | None]: Pulse schedule and its corresponding final layout (Initial Re-mapping + SWAPs routing) of
                the Original Logical Qubits (l_q) in the physical circuit (wires): [l_q in wire 0, l_q in wire 1, ...] (None = trivial mapping).
        """
        # Default values:
        if transpilation_config is None:
            transpilation_config = DigitalTranspilationConfig()

        # Unpack dataclass attributes:
        routing, placer, router, routing_iterations, optimize = transpilation_config._attributes_ordered

        # Routing stage;
        if routing:
            circuit_gates, nqubits, final_layout = self.route_circuit(circuit, placer, router, routing_iterations)
        else:
            circuit_gates, nqubits = circuit.queue, circuit.nqubits
            final_layout = None  # Random mapping

        # Optimze qibo gates, cancelling redundant gates:
        if optimize:
            circuit_gates = self.optimize_gates(circuit_gates)

        # Unroll to Natives gates:
        circuit_gates = self.gates_to_native(circuit_gates)

        # Add phases from RZs and CZs to Drags:
        circuit_gates = self.add_phases_from_RZs_and_CZs_to_drags(circuit_gates, nqubits)

        # Optimze transpiled qibo gates, cancelling redundant gates:
        if optimize:
            circuit_gates = self.optimize_transpiled_gates(circuit_gates)

        # Pulse schedule stage:
        pulse_schedule = self.gates_to_pulses(circuit_gates)

        return pulse_schedule, final_layout

    def route_circuit(
        self,
        circuit: Circuit,
        placer: Placer | type[Placer] | tuple[type[Placer], dict] | None = None,
        router: Router | type[Router] | tuple[type[Router], dict] | None = None,
        iterations: int = 10,
        coupling_map: tuple[int, int] | None = None,
    ) -> tuple[list[gates.Gate], int, list[int]]:
        """Routes the virtual/logical qubits of a circuit to the physical qubits of a chip. Returns and logs the final qubit layout.

        This process uses the provided ``placer``, ``router``, and ``routing_iterations`` parameters if they are passed; otherwise, default values are applied.

        **Examples:**

        If we instantiate some ``Circuit``, ``Platform`` and ``CircuitTranspiler`` objects like:

        .. code-block:: python

            from qibo import gates
            from qibo.models import Circuit
            from qibo.transpiler.placer import ReverseTraversal, Random
            from qibo.transpiler.router import Sabre
            from qililab import build_platform
            from qililab.digital import CircuitTranspiler

            # Create circuit:
            c = Circuit(5)
            c.add(gates.CNOT(1, 0))

            # Create platform:
            platform = build_platform(runcard="<path_to_runcard>")
            coupling_map = platform.digital_compilation_settings.topology

            # Create transpiler:
            transpiler = CircuitTranspiler(platform)

        Now we can transpile like:

        .. code-block:: python

            # Default Transpilation:
            routed_circuit, qubits, final_layouts = transpiler.route_circuit(c)  # Defaults to ReverseTraversal, Sabre and platform connectivity

            # Non-Default Random placer, and coupling_map specified:
            routed_circuit, qubits, final_layouts = transpiler.route_circuit(c, placer=Random, router=Sabre, coupling_map)

            # Specifying one of the a kwargs:
            routed_circuit, qubits, final_layouts = transpiler.route_circuit(c, placer=Random, router=(Sabre, {"lookahead": 2}))

        Args:
            circuit (Circuit): circuit to route.
            placer (Placer | type[Placer] | tuple[type[Placer], dict], optional): ``Placer`` instance, or subclass ``type[Placer]`` to
                use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``ReverseTraversal``.
            router (Router | type[Router] | tuple[type[Router], dict], optional): ``Router`` instance, or subclass ``type[Router]`` to
                use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``Sabre``.
            iterations (int, optional): Number of times to repeat the routing pipeline, to keep the best stochastic result. Defaults to 10.
            coupling_map (tuple[int, int], optional): coupling map of the chip to route. This topology will be the one that rules,
                which will overwrite any other in an instance of router or placer. Defaults to the platform topology.

        Returns:
            tuple[list[Gate], int, list[int]]: List of gates of the routed circuit, number of qubits in it, and its corresponding final layout (Initial
                Re-mapping + SWAPs routing) of the Original Logical Qubits (l_q) in the physical circuit (wires): [l_q in wire 0, l_q in wire 1, ...].

        Raises:
            ValueError: If StarConnectivity Placer and Router are used with non-star topologies.
        """
        # Get the chip's connectivity
        topology = nx.Graph(coupling_map if coupling_map is not None else self.settings.topology)
        circuit_router = CircuitRouter(topology, placer, router)

        circuit, final_layout = circuit_router.route(circuit, iterations)

        return circuit.queue, circuit.nqubits, final_layout

    @staticmethod
    def optimize_gates(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Main method to optimize the gates of a Quantum Circuit before unrolling to native gates.

        Currently only applies a cancellation for adjacent hermitian gates (H, X, Y, Z, CNOT, CZ, SWAP).

        The total optimization can/might be expanded in the future to include more complex gate optimization.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the Qibo circuit to optimize.

        Returns:
            list[gates.Gate]: list of the gates of the Qibo circuit, optimized.
        """
        return CircuitOptimizer.optimize_gates(circuit_gates)

    @staticmethod
    def gates_to_native(circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Maps Qibo gates to a hardware native implementation (CZ, RZ, Drag, Wait and M (Measurement))
            - CZ gates are our 2 qubit gates
            - RZ gates are applied as virtual Z gates if optimize=True in the transpiler
            - Drag gates are our single qubit gates
            - Wait gates add wait time at a single qubit
            - Measurement gates measure the circuit

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the Qibo circuit, to pass to native.

        Returns:
            list[gates.Gate]: list of native gates of the Qibo circuit.

        """
        return translate_gates(circuit_gates)

    def add_phases_from_RZs_and_CZs_to_drags(self, circuit_gates: list[gates.Gate], nqubits: int) -> list[gates.Gate]:
        """This method adds the phases from RZs and CZs gates of the circuit to the next Drag gates.

            - The CZs added phases on the Drags, come from a correction from their calibration, stored on the setting of the CZs.
            - The RZs added phases on the Drags, come from commuting all the RZs all the way to the end of the circuit, so they can be deleted as "virtual Z gates".

        This is done by moving all RZ to the left of all operators as a single RZ. The corresponding cumulative rotation
        from each RZ is carried on as phase in all drag pulses left of the RZ operator.

        Virtual Z gates are also applied to correct phase errors from CZ gates.

        The final RZ operator left to be applied as the last operator in the circuit can afterwards be removed since the last
        operation is going to be a measurement, which is performed on the Z basis and is therefore invariant under rotations
        around the Z axis.

        This last step can also be seen from the fact that an RZ operator applied on a single qubit, with no operations carried
        on afterwards induces a phase rotation. Since phase is an imaginary unitary component, its absolute value will be 1
        independent on any (unitary) operations carried on it.

        Mind that moving an operator to the left is equivalent to applying this operator last so
        it is actually moved to the _right_ of ``Circuit.queue`` (last element of list).

        For more information on virtual Z gates, see https://arxiv.org/abs/1612.00858

        Args:
            circuit_gates (list[gates.Gate]): list of native gates of the circuit, to pass phases to the Drag gates.
            nqubits (int): Number of qubits of the circuit.

        Returns:
            list[gates.Gate]: list of native gates of the circuit, with phases passed to the Drag gates.
        """
        return self.optimizer.add_phases_from_RZs_and_CZs_to_drags(circuit_gates, nqubits)

    def optimize_transpiled_gates(self, circuit_gates: list[gates.Gate]) -> list[gates.Gate]:
        """Main method to optimize the gates of a Quantum Circuit after having unrolled to native gates.

        Currently only bunches consecutive Drag gates, with same phi's, together into a single one, and removes redundant Drag gates.

        The total optimization can/might be expanded in the future to include more complex optimizations.

        Args:
            circuit_gates (list[gates.Gate]): list of gates of the transpiled circuit, to optimize.

        Returns:
            list[gates.Gate]: list of gates of the transpiled circuit, optimized.
        """
        return self.optimizer.optimize_transpiled_gates(circuit_gates)

    def gates_to_pulses(self, circuit_gates: list[gates.Gate]) -> PulseSchedule:
        """Translates a Qibo circuit into its corresponding pulse sequences.

        For each circuit gate we look up for its corresponding gates settings in the runcard (the name of the class of the circuit
        gate and the name of the gate in the runcard should match) and load its schedule of GateEvents.

        Each gate event corresponds to a concrete pulse applied at a certain time w.r.t the gate's start time and through a specific bus
        (see gates settings docstrings for more details).

        Measurement gates are handled in a slightly different manner. For a circuit gate M(0,1,2) the settings for each M(0), M(1), M(2)
        will be looked up and will be applied in sync. Note that thus a circuit gate for M(0,1,2) is different from the circuit sequence
        M(0)M(1)M(2) since the later will not be necessarily applied at the same time for all the qubits involved.

        Times for each qubit are kept track of with the dictionary ``time``.

        The times at which each pulse is applied are padded if they are not multiples of the minimum clock time. This means that if min clock
        time is 4 and a pulse applied to qubit k lasts 17ns, the next pulse at qubit k will be at t=20ns

        Args:
            circuit_gates (list[gates.Gate]): list of native gates of the Qibo circuit.

        Returns:
            PulseSequences: equivalent :class:`PulseSequences` class.
        """
        circuit_to_pulses = CircuitToPulses(self.settings)

        return circuit_to_pulses.run(circuit_gates)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np
import plotly.graph_objects as go


class QbloxDraw:
    def _call_handlers(self, program_line, param, register, data_draw, waveform_seq):
        """Calls the handlers.

        Args:
            program_line (tuple): line of the Q1ASM program parsed.
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).
            register (dictionary): registers of the Q1ASM.
            data_draw (list): nested list for each waveform, data points until current time.
            waveform_seq (Dict items): waveform data as provided by the sequencer
        """

        action_type = program_line[0]
        if action_type == "set_freq":
            param = self._handle_freq_phase_draw(program_line, param, register, "intermediate_frequency")
        elif action_type == "set_ph":
            param = self._handle_freq_phase_draw(program_line, param, register, "phase")
        elif action_type == "reset_ph":
            param = self._handle_reset_phase_draw(param)
        elif action_type == "set_awg_offs":
            param = self._handle_offset(program_line, param)
        elif action_type == "set_awg_gain":
            param = self._handle_gain_draw(program_line, param, register)
        elif action_type == "play":
            data_draw = self._handle_play_draw(data_draw, program_line, waveform_seq, param)
        elif action_type == "wait":
            data_draw = self._handle_wait_draw(data_draw, program_line, param)
        elif action_type == "upd_param":
            data_draw = self._handle_wait_draw(data_draw, program_line, param)
        elif action_type == "add":
            self._handle_add_draw(register, program_line)
        return param, register, data_draw

    def _calculate_scaling_and_offsets(self, param, i_or_q):
        if i_or_q == "I":
            scaling_factor, max_voltage = self._get_scaling_factors(
                param, param.get("q1asm_offset_i", 0), param.get("ac_offset_i", 0)
            )
            gain = param.get("gain_i", 1)
        elif i_or_q == "Q":
            scaling_factor, max_voltage = self._get_scaling_factors(
                param, param.get("q1asm_offset_q", 0), param.get("ac_offset_q", 0)
            )
            gain = param.get("gain_q", 1)
        return scaling_factor, max_voltage, gain

    def _get_scaling_factors(self, param, dynamic_off, ac_off):
        if param["hardware_modulation"]:
            return (
                (param["max_voltage"] / (np.sqrt(2)), param["max_voltage"] / (np.sqrt(2)))
                if dynamic_off == 0 and ac_off == 0
                else (param["max_voltage"] / (np.sqrt(2)), param["max_voltage"])
            )
        return (param["max_voltage"], param["max_voltage"])  # (scaling factor, max voltage)

    def _handle_play_draw(self, data_draw, program_line, waveform_seq, param):
        """Play a waveform by appending the stored data list.

        Args:
            data_draw (list): nested list for each waveform, data points until current time.
            program_line (tuple): line of the Q1ASM program parsed with a play instruction.
            waveform_seq (Dict items): waveform data as provided by the sequencer
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).

        Returns:
            Appended data_draw considering the play, the gain and the offset.
            And appended the IF and phase keys of the param dictionary. Each of these is a np array where 1 data point represents 1 ns (same as the waveforms).
        """
        output_path1, output_path2, _ = map(int, program_line[1].split(","))
        # modify and store the wf data
        for idx, output_path in enumerate([output_path1, output_path2]):
            for _, waveform_value in waveform_seq:
                index = waveform_value["index"]
                if index == output_path:
                    iq = "I" if idx == 0 else "Q"
                    scaling_factor, max_voltage, gain = self._calculate_scaling_and_offsets(param, iq)
                    scaled_array = np.array(waveform_value["data"]) * scaling_factor
                    modified_waveform = np.clip(scaled_array * gain, -max_voltage, max_voltage)
                    data_draw[idx] = np.append(data_draw[idx], modified_waveform)

        # extend the IF and phase by the length of the wf
        for key in ["intermediate_frequency", "phase", "q1asm_offset_i", "q1asm_offset_q"]:
            if param.get(f"{key}_new", False) is True:
                param[key].extend([param[key][-1]] * (len(scaled_array) - 1))
                param[f"{key}_new"] = False
            else:
                param[key].extend([param[key][-1]] * len(scaled_array))
        return data_draw

    def _handle_wait_draw(self, data_draw, program_line, param):
        """Play a wwait by appending the stored data list.

        Args:
            data_draw (list): nested list for each waveform, data points until current time.
            program_line (tuple): line of the Q1ASM program parsed with a wait instruction.
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).

        Returns:
            Appended data_draw considering the wait and the offset.
            And appended the IF and phase keys of the param dictionary. Each of these is a np array where 1 data point represents 1 ns (same as the waveforms).
        """

        y_wait = np.linspace(0, 0, int(program_line[1]))
        data_draw[0] = np.append(data_draw[0], (y_wait))
        data_draw[1] = np.append(data_draw[1], (y_wait))

        # extend the IF and phase by the length of the wf
        for key in ["intermediate_frequency", "phase", "q1asm_offset_i", "q1asm_offset_q"]:
            if param.get(f"{key}_new", False) is True:
                param[key].extend([param[key][-1]] * (len(y_wait) - 1))
                param[f"{key}_new"] = False
            else:
                param[key].extend([param[key][-1]] * len(y_wait))
        return data_draw

    def _handle_gain_draw(self, program_line, param, register):
        """Updates the param dictionary when a gain is set in the program

        Args:
            data_draw (list): nested list for each waveform, data points until current time.
            program_line (tuple): line of the Q1ASM program parsed with a gain or offset instruction.
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).

        Returns:
            Appends the param dictionary
        """
        i_val, q_val = program_line[1].split(", ")
        i_val = float(self._get_value(i_val, register)) / 32767
        q_val = float(self._get_value(q_val, register)) / 32767
        param["gain_i"], param["gain_q"] = i_val, q_val
        return param

    def _handle_offset(self, program_line, param):
        """Updates the param dictionary when an offset is set in the program

        Args:
            data_draw (list): nested list for each waveform, data points until current time.
            program_line (tuple): line of the Q1ASM program parsed with an offset instruction.
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).

        Returns:
            Appends the param dictionary
        """
        offi, offq = map(int, program_line[1].split(","))

        if param["hardware_modulation"]:
            scaling_offset = param["max_voltage"] / np.sqrt(2)
        else:
            scaling_offset = param["max_voltage"]

        for x, off in zip(["q1asm_offset_i", "q1asm_offset_q"], [offi, offq]):
            new_key = f"{x}_new"
            if param[new_key]:
                param[x][-1] = off * scaling_offset / 32767
            else:
                param[x].append(off * scaling_offset / 32767)
            param[new_key] = True
        return param

    def _handle_freq_phase_draw(self, program_line, param, register, key):
        """Updates the param dictionary when a freq or phase is set in the program

        Args:
            data_draw (list): nested list for each waveform, data points until current time.
            program_line (tuple): line of the Q1ASM program parsed with a freq or phase instruction.
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).
            key (string): the key is freq or phase.

        Returns:
            Appends the param dictionary.
        """
        new_key = f"{key}_new"
        if param[new_key]:
            param[key][-1] = self._get_value(program_line[1], register)
        else:
            param[key].append(self._get_value(program_line[1], register))
        param[new_key] = True
        return param

    def _handle_reset_phase_draw(self, param):
        """Handles a reset phase

        Args:
            param (dictionary): parameters of the bus (IF, phase, offset, hardware modulation).

        Returns:
            Sets the phase as 0 in the param dictionary.
        """
        if param["phase_new"]:
            param["phase"][-1] = 0
        else:
            param["phase"].append(0)
        param["phase_new"] = True
        return param

    def _handle_add_draw(self, register, program_line):
        """Updates the register dictionary when a line of the parsed program has a move command.

        Args:
            register (dictionary): registers of the Q1ASM.
            program_line (tuple): line of the Q1ASM program parsed with a freq or phase instruction.

        Returns:
            Updated register.
        """
        a, b, destination = program_line[1].split(", ")
        register[destination] = self._get_value(a, register) + self._get_value(b, register)
        return register

    def _get_value(self, x, register):
        """
        Args:
            x (string|int|float): desired value or variable.
            register (dictionary): registers of the Q1ASM.

        Returns:
            If the input is a number it returns a float; if the input is a key of the registery it returns th evalue as a float
        """
        if x is not None:
            if x.isdigit():
                return float(x)
            if x in register:
                return float(register[x])
        return None

    def _parse_program(self, sequences):
        """Parses the program dictionary of the sequence.
        Args:
            sequences (dictionary): for each bus, it is made up of nested dictionaries (waveforms,weights,acquisitions and program)

        Returns:
            Returns seq_parsed_program with each section (setup/main) as a list of tuple (instruction, numerical value, (tuple of loop label), index)
            Example: Q1ASM has play 0,1,30 -  the equivalent seq_parsed_program is (play,'0,1,30', (),1)
                where 1 is the index and () is a tuple of the loop labels
                ie: (avg_0, loop_0) meaning the current isntruction is part of 2 loops, avg_0 as top and loop_0 as nested.

        """
        seq_parsed_program = {}
        for bus in sequences:  # Iterate through the bus of the sequences
            sequence = sequences[bus].todict()
            program_line = sequence["program"].split("\n")
            processed_lines = []

            # parse all lines of qprogram and deal with special case of reset phase
            for line in program_line:
                stripped_line = line.strip()
                if stripped_line.startswith("reset_ph"):
                    processed_lines.append("reset_ph             2")
                elif stripped_line.startswith("nop"):
                    processed_lines.append("nop             2")
                else:
                    processed_lines.append(stripped_line)
            line_data = "\n".join(processed_lines)
            pattern = r"(\w+):|(\w+)(?:\s+([^\n]*))?"
            matches = re.findall(pattern, line_data)  # (section/loop, instruction, numerical value)

            program_parsed = {
                "setup": [],
                "main": [],
            }  # each section is a list of tuples (instruction, numerical value, (tuple of loop label), index)
            index = 0  # Track execution order
            # Create a list of the program section made of tuples (command, args, index)
            loop_label = ()
            for section, instruction, numerical_value in matches:
                if section:
                    if section == "setup":
                        current_section = section
                        index = 0
                    if section == "main":
                        current_section = section
                        index = 0
                    elif section not in ["setup", "main"]:
                        loop_label += (section,)  # add the label of the loop
                elif instruction:
                    numerical_value = numerical_value if numerical_value else ""
                    program_parsed[current_section].append(
                        (instruction, numerical_value, loop_label, index)
                    )  # Store index
                    index += 1  # Increment order index
                    for la in loop_label:  # remove the @, eg: Q1ASM has @loop_0 and we want loop_0
                        if ("@" + la) in numerical_value:
                            loop_label = tuple(l for l in loop_label if l != la)
                sequence["program"] = program_parsed
            del sequence[
                "program"
            ][
                "main"
            ][
                -3:
            ]  # delete the last 3 lines of the Q1ASM that are always hardcoded - tempers with the _new flag later on if not removed here
            seq_parsed_program[bus] = sequence
        return seq_parsed_program

    def draw(self, sequencer, runcard_data=None, averages_displayed=False) -> dict:
        """Parses the program dictionary of the sequence, plots the waveforms and generates waveform data.
        Args:
            sequencer: The compiled qprogram, either at the platform or qprogram level.
            runcard_data (dictionary): parameters of the bus (IF, phase, offset, hardware modulation) retrieved from the runcard if using the platform
                This gets renamed as param to overwrite/add values from the sequencer.
            averages_displayed (bool): Determines how looping variables are handled. If False (default), all loops
                on the sequencer starting with `avg_` will only iterate once. If True, all iterations are displayed.
            loops_displayed (bool): Determines how looping variables are handled. If False (default), all loops
                on the sequencer starting with `loop_` will only iterate once. If True, all iterations are displayed.


        Returns:
            data_draw (dictionary): A dictionary where keys are bus aliases and values are lists containing numpy arrays for
                the I and Q components. This includes all data points used for plotting the waveforms.

        Note:
            This function also **plots** the waveforms using the generated data.

        """
        Q1ASM_ordered = self._parse_program(
            sequencer.sequences.copy()
        )  # (instruction, value, label of the loops, index)
        data_draw = {}
        parameters = {}
        for bus, _ in Q1ASM_ordered.items():
            # Load data from the runcard or create the keys of the dict containing all paraemters (frequency, phase, offsets, gains, etc...)
            if runcard_data is not None:
                parameters[bus] = {
                    key: runcard_data[bus][key] for key in runcard_data[bus]
                }  # retrieve runcard data if the qblox draw is called when a platform has been built
                IF = parameters[bus]["intermediate_frequency"] * 4
                parameters[bus]["intermediate_frequency"] = [IF]
                parameters[bus]["ac_offset_i"] = parameters[bus]["offset_i"]
                parameters[bus]["ac_offset_q"] = parameters[bus]["offset_q"]
                if parameters[bus]["instrument_name"] in {"QCM", "QCM-RF"}:
                    parameters[bus]["max_voltage"] = 2.5
                elif parameters[bus]["instrument_name"] in {"QRM", "QRM-RF"}:
                    parameters[bus]["max_voltage"] = 0.5
            else:  # no runcard uploaded- running qp directly
                parameters[bus] = {}
                parameters[bus]["intermediate_frequency"] = [0]
                parameters[bus]["ac_offset_i"] = [0]
                parameters[bus]["ac_offset_q"] = [0]
                parameters[bus]["dac_offset_i"] = [0]
                parameters[bus]["dac_offset_q"] = [0]
                parameters[bus]["hardware_modulation"] = True  # if plotting directly from qp, plot i and q
                parameters[bus]["max_voltage"] = 1
                parameters[bus]["instrument_name"] = "QProgram"
            parameters[bus]["q1asm_offset_i"] = [0]
            parameters[bus]["q1asm_offset_q"] = [0]
            parameters[bus]["phase"] = [0]

            # flags to determine if the phase or freq has been updated
            parameters[bus]["phase_new"] = True
            parameters[bus]["intermediate_frequency_new"] = True
            parameters[bus]["q1asm_offset_i_new"] = True
            parameters[bus]["q1asm_offset_q_new"] = True

            wf1: list[float] = []
            wf2: list[float] = []
            param = parameters[bus]
            instructions_ran = []  # keep track of the instructions that have been done
            data_draw[bus] = [wf1, wf2]
            label_done = []  # list to keep track of the label once they have been looped over
            wf = Q1ASM_ordered[bus]["waveforms"].items()

            # Loop through the program to store the register and have the information on the loops
            register = {}
            register["avg_no_loop"] = 1
            loop_info = {}
            for Q1ASM_line in Q1ASM_ordered[bus]["program"]["main"]:
                if Q1ASM_line[0] == "move":
                    reg = Q1ASM_line[1].split(",")[1].strip()
                    value = Q1ASM_line[1].split(",")[0].strip()
                    register[reg] = int(value)

                # sorted labels (label, [start index, end index, register key])
                _, value, label, index = Q1ASM_line
                for l in label:
                    if l not in loop_info:
                        loop_info[l] = [index, index, None]
                    else:
                        loop_info[l][1] = index
                        loop_info[l][2] = value.split(",")[0]
                        if l.startswith("avg") and not averages_displayed:
                            loop_info[l][2] = "avg_no_loop"
                sorted_labels = sorted(loop_info.items(), key=lambda x: x[1][0])

            for Q1ASM_line in Q1ASM_ordered[bus]["program"]["main"]:

                def process_loop(recursive_input, i):
                    (label, [start, end, value]) = recursive_input
                    if label not in label_done:
                        label_done.append(label)
                    for x in range(register[value], 0, -1):
                        current_idx = start
                        while current_idx <= end:
                            item = Q1ASM_ordered[bus]["program"]["main"][current_idx]
                            wf = Q1ASM_ordered[bus]["waveforms"].items()
                            _, value, label, _ = item
                            for la in label:
                                if la not in label_done:  # nested loop
                                    new_label = la
                                    result = next(
                                        (element for element in sorted_labels if element[0] == new_label), None
                                    )  # retrieve the start/end/variable of the new label
                                    current_idx = process_loop(result, current_idx)
                                    label_index = {}
                                    # check if there is a nested loop, if yes need to remove it from label_dne, otherwise it wont loop over in the next iteration of the parent
                                    for a in label_done:
                                        element = next((e for e in sorted_labels if e[0] == a), None)[1][0]
                                        label_index[a] = int(element)
                                    max_value = max(label_index.values())
                                    max_keys = [key for key, value in label_index.items() if value == max_value]
                                    label_done.remove(max_keys[0])

                            item = Q1ASM_ordered[bus]["program"]["main"][current_idx]
                            wf = Q1ASM_ordered[bus]["waveforms"].items()
                            instructions_ran.append(item[-1])
                            self._call_handlers(item, param, register, data_draw[bus], wf)
                            current_idx += 1
                    return current_idx

                if (
                    Q1ASM_line[2] and Q1ASM_line[-1] not in instructions_ran
                ):  # if there is a loop label and if the index has not been ran before
                    index = 0
                    for x in sorted_labels:
                        if x[0] == Q1ASM_line[2][0]:
                            input_recursive = x
                            break  # Stop as soon as the first match is found
                    process_loop(input_recursive, index)

                elif Q1ASM_line[-1] not in instructions_ran:  # run if no loop label
                    self._call_handlers(Q1ASM_line, param, register, data_draw[bus], wf)

            # remove the latest freq, phase and offset data points if nothing played after
            for key in ["intermediate_frequency", "phase", "q1asm_offset_i", "q1asm_offset_q"]:
                if param[f"{key}_new"]:
                    param[key].pop()
            parameters[bus] = param

        data_draw = self._oscilloscope_plotting(data_draw, parameters)
        return data_draw

    def _oscilloscope_plotting(self, data_draw, parameters):
        """Plots the waveform data and applies the phase and frequency np array to the data
        Args:
            parameters (dictionary): parameters of all buses (IF, phase, offset, hardware modulation).
            data_draw (dictionary): A dictionary where keys are bus aliases and values are lists containing numpy arrays for
                the I and Q components. This includes all data points used for plotting the waveforms.

        Returns:
            data_draw (dictionary): A dictionary where keys are bus aliases and values are lists containing numpy arrays for
                the I and Q components. This includes all data points used for plotting the waveforms. This function modifies this dictionary,
                it adds the offsets, the phase and the frequency to the waveforms.

        Note:
            This function also **plots** the waveforms using the generated data.
        """

        data_keys = list(data_draw.keys())

        fig = go.Figure()

        for idx, key in enumerate(data_keys):
            q1asm_offset_i = np.array(parameters[key]["q1asm_offset_i"])
            q1asm_offset_q = np.array(parameters[key]["q1asm_offset_q"])
            volt_bounds = parameters[key]["max_voltage"]
            dac_offset_i, dac_offset_q = parameters[key]["dac_offset_i"], parameters[key]["dac_offset_q"]
            ac_offset_i, ac_offset_q = (
                parameters[key]["ac_offset_i"] * volt_bounds / np.sqrt(2),
                parameters[key]["ac_offset_q"] * volt_bounds / np.sqrt(2),
            )
            if not parameters[key]["hardware_modulation"]:  # if hardware modulation is disabled, do not plot Q
                ac_offset_i = ac_offset_i * volt_bounds
                waveform_flux = np.clip(
                    (np.array(data_draw[key][0]) + q1asm_offset_i + ac_offset_i + dac_offset_i),
                    -volt_bounds,
                    volt_bounds,
                )
                data_draw[key][0] = waveform_flux
                data_draw[key][1] = None
                fig.add_trace(go.Scatter(y=waveform_flux, mode="lines", name="Flux"))
            else:
                ac_offset_i, ac_offset_q = (
                    ac_offset_i * volt_bounds / np.sqrt(2),
                    ac_offset_q * volt_bounds / np.sqrt(2),
                )
                wf1, wf2 = data_draw[key][0], data_draw[key][1]
                fs = 1e9  # sampling frequency of the qblox
                t = np.arange(0, len(wf1)) / fs

                # make freq and phase np arrays and convert from qblox values
                freq = np.array(parameters[key]["intermediate_frequency"]) / 4
                phase = np.array(parameters[key]["phase"]) * (2 * np.pi / 1e9)

                cos_term = np.cos(2 * np.pi * freq * t + phase)
                sin_term = np.sin(2 * np.pi * freq * t + phase)

                # Add the offsets to the waveforms and ensure it is in the voltage range of the instrument
                wf1_offsetted = np.clip((np.array(wf1) + q1asm_offset_i), -volt_bounds, volt_bounds)
                wf2_offsetted = np.clip((np.array(wf2) + q1asm_offset_q), -volt_bounds, volt_bounds)
                path0 = (
                    (cos_term * np.array(wf1_offsetted) - sin_term * np.array(wf2_offsetted))
                    + dac_offset_i
                    + ac_offset_i
                )
                path1 = (
                    (sin_term * np.array(wf1_offsetted) + cos_term * np.array(wf2_offsetted))
                    + dac_offset_q
                    + ac_offset_q
                )

                # clip the final signal
                path0_clipped = np.clip(path0, -volt_bounds, volt_bounds)
                path1_clipped = np.clip(path1, -volt_bounds, volt_bounds)

                data_draw[key][0], data_draw[key][1] = path0_clipped, path1_clipped

                fig.add_trace(go.Scatter(y=path0_clipped, mode="lines", name=f"{key} I"))
                fig.add_trace(go.Scatter(y=path1_clipped, mode="lines", name=f"{key} Q"))

        if parameters[key]["instrument_name"] == "QProgram":
            fig.update_yaxes(title_text="Amplitude [a.u.]")
        else:
            fig.update_yaxes(title_text="Voltage [V]")
        fig.update_xaxes(title_text="Time [ns]")

        # Update layout
        fig.update_layout(
            height=200 * len(data_keys),
            width=1100,
            title_text="QBlox Oscillator simulation",
            showlegend=True,
        )
        fig.show()

        return data_draw



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# mypy: disable-error-code="union-attr, arg-type"
import inspect
import os
import threading
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from time import perf_counter
from types import LambdaType
from typing import TYPE_CHECKING, Callable
from uuid import UUID

import numpy as np
from rich.progress import BarColumn, Progress, TaskID, TextColumn, TimeElapsedColumn

from qililab.qprogram.blocks import Average, Block, ForLoop, Loop, Parallel
from qililab.qprogram.experiment import Experiment
from qililab.qprogram.operations import ExecuteQProgram, GetParameter, Measure, Operation, SetParameter
from qililab.qprogram.operations.set_crosstalk import SetCrosstalk
from qililab.qprogram.variable import Variable
from qililab.result.experiment_results_writer import (
    ExperimentMetadata,
    ExperimentResultsWriter,
    MeasurementMetadata,
    QProgramMetadata,
    VariableMetadata,
)
from qililab.result.qprogram.qprogram_results import QProgramResults
from qililab.utils.serialization import serialize

if TYPE_CHECKING:
    from qililab.platform.platform import Platform


@dataclass
class VariableInfo:
    """Dataclass to store information of a Variable"""

    uuid: UUID
    label: str
    values: np.ndarray


class ExperimentExecutor:
    """Manages the execution of a quantum experiment.

    The `ExperimentExecutor` class is responsible for orchestrating the execution of a quantum experiment on a given platform. It traverses the experiment's structure, handles loops and operations, and stores the results in real-time to ensure data integrity even in case of interruptions.

    Key responsibilities include:

    - Preparing metadata and loop structures before execution.
    - Managing variables and their scopes within loops and blocks.
    - Executing operations in the correct sequence with proper parameter settings.
    - Streaming results to an HDF5 file using `ExperimentResultsWriter`.

    This class provides a high-level interface to execute complex experiments involving nested loops, parameter sweeps, and qprogram executions, while efficiently managing resources and progress tracking.

    Args:
        platform (Platform): The platform on which the experiment is to be executed.
        experiment (Experiment): The experiment object defining the sequence of operations and loops.
        base_data_path (str): The base directory path where the experiment results will be stored.

    Example:
        .. code-block::

            from qililab.data_management import build_platform
            from qililab.qprogram import Experiment
            from qililab.executor import ExperimentExecutor

            # Initialize the platform
            platform = build_platform(runcard="path/to/runcard.yml")

            # Define your experiment
            experiment = Experiment(label="my_experiment")
            # Add blocks, loops, operations to the experiment
            # ...

            # Set the base data path for storing results
            base_data_path = "/data/experiments"

            # Create the ExperimentExecutor
            executor = ExperimentExecutor(platform=platform, experiment=experiment, base_data_path=base_data_path)

            # Execute the experiment
            results_path = executor.execute()
            print(f"Results saved to {results_path}")

    Note:
        - Ensure that the platform and experiment are properly configured before execution.
        - The results will be saved in a timestamped directory within the `base_data_path`.
    """

    def __init__(self, platform: "Platform", experiment: Experiment):
        self.platform = platform
        self.experiment = experiment

        # Registry of all variables used in the experiment with their labels and values
        self._all_variables: dict = defaultdict(lambda: {"label": None, "values": {}})

        # Mapping from each Block to the list of variables associated with that block
        self._variables_per_block: dict[Block, list[VariableInfo]] = {}

        # Mapping from each ExecuteQProgram operation to its execution index (order of execution)
        self._qprogram_execution_indices: dict[ExecuteQProgram, int] = {}

        # Variables that uses flux for further processing and saving the right bias
        # TODO: implement a way to save the bias based on the same principle as HW loops Xtalk
        self._flux_variables: dict[str, np.ndarray] = {}

        # Stack to keep track of variables in the experiment context (outside QPrograms)
        self._experiment_variables_stack: list[list[VariableInfo]] = []

        # Stack to keep track of variables within QPrograms
        self._qprogram_variables_stack: list[list[VariableInfo]] = []

        # Counter for the number of QPrograms encountered.
        self._qprogram_index = 0

        # Counter for the number of measurements within a QProgram.
        self._measurement_index = 0

        # Number of shots for averaging measurements. It is updated when an Average block is encountered.
        self._shots = 1

        # Metadata dictionary containing information about the experiment structure and variables.
        self._metadata: ExperimentMetadata

        # ExperimentResultsWriter object responsible for saving experiment results to file in real-time.
        self._results_writer: ExperimentResultsWriter

    def _prepare_metadata(self, executed_at: datetime):
        """Prepares the loop values and result shape before execution."""

        def traverse_experiment(block: Block):
            """Traverses the blocks to gather loop information and determine result shape."""
            if isinstance(block, (Loop, ForLoop, Parallel)):
                variables = self._get_variables_of_loop(block)

                self._experiment_variables_stack.append(variables)

            # Recursively traverse nested blocks or loops
            for element in block.elements:
                if isinstance(element, Block):
                    traverse_experiment(element)
                # Handle ExecuteQProgram operations and traverse their loops
                if isinstance(element, ExecuteQProgram):
                    if isinstance(element.qprogram, LambdaType):
                        signature = inspect.signature(element.qprogram)
                        call_parameters = {param.name: 0 for param in signature.parameters.values()}
                        qprogram = element.qprogram(**call_parameters)
                        traverse_qprogram(qprogram.body)
                    else:
                        traverse_qprogram(element.qprogram.body)
                    self._qprogram_execution_indices[element] = self._qprogram_index
                    self._measurement_index = 0
                    self._qprogram_index += 1

            if isinstance(block, (Loop, ForLoop, Parallel)):
                del self._experiment_variables_stack[-1]

        def traverse_qprogram(block: Block):
            """Traverses a QProgram to gather loop information."""
            if isinstance(block, (Loop, ForLoop, Parallel)):
                variables = self._get_variables_of_loop(block)

                self._qprogram_variables_stack.append(variables)
            if isinstance(block, Average):
                self._shots = block.shots

            # Recursively handle nested blocks within the QProgram
            for element in block.elements:
                if isinstance(element, Block):
                    traverse_qprogram(element)
                if isinstance(element, Measure):
                    finalize_measurement_structure()

            if isinstance(block, (Loop, ForLoop, Parallel)):
                del self._qprogram_variables_stack[-1]
            if isinstance(block, Average):
                self._shots = 1

        def finalize_measurement_structure():
            """Finalize the structure of a measurement when a Measure operation is encountered."""
            qprogram_name = f"QProgram_{self._qprogram_index}"
            measurement_name = f"Measurement_{self._measurement_index}"

            # Ensure QProgram exists in the structure
            if qprogram_name not in self._metadata["qprograms"]:
                self._metadata["qprograms"][qprogram_name] = QProgramMetadata(
                    variables=[
                        VariableMetadata(label=variable.label, values=variable.values)
                        for sublist in self._experiment_variables_stack
                        for variable in sublist
                    ],
                    dims=[[variable.label for variable in sublist] for sublist in self._experiment_variables_stack],
                    measurements={},
                )

            # Add QProgram loops and the measurement
            self._metadata["qprograms"][qprogram_name]["measurements"][measurement_name] = MeasurementMetadata(
                variables=[
                    VariableMetadata(label=variable.label, values=variable.values)
                    for sublist in self._qprogram_variables_stack
                    for variable in sublist
                ],
                dims=[[variable.label for variable in sublist] for sublist in self._qprogram_variables_stack],
                shape=(
                    *tuple(
                        len(sublist[0].values)
                        for sublist in self._experiment_variables_stack + self._qprogram_variables_stack
                    ),
                    2,
                ),
                shots=self._shots,
            )

            # Increase index of measurements
            self._measurement_index += 1

        self._metadata = ExperimentMetadata(
            platform=serialize(self.platform.to_dict()),
            experiment=serialize(self.experiment),
            executed_at=executed_at,
            execution_time=0.0,
            qprograms={},
        )
        traverse_experiment(self.experiment.body)
        self._all_variables = dict(self._all_variables)

    def _prepare_operations(self, block: Block, progress: Progress):
        """Traverse blocks, store generated Python functions, and return the stored operations."""

        # A mapping from block UUID to the associated Progress TaskID
        task_ids: dict[UUID, TaskID] = {}

        # A list of operations to execute
        operations: list[Callable] = []

        # A mapping from block UUID to the index of the current value of its variable
        loop_indices: dict[Block, int] = {}

        # A mapping from variable UUID to current value of the variable
        current_value_of_variable: dict[UUID, int | float] = {}

        def handle_loop(block: ForLoop | Loop | Parallel) -> list[Callable]:
            """Common logic for handling ForLoop and Loop blocks."""
            loop_operations: list[Callable] = []

            # Determine loop parameters based on the type of block
            label = ",".join([variable.label for variable in self._variables_per_block[block]])
            shape = self._variables_per_block[block][0].values.shape[-1]

            # Create the progress bar for the loop
            def create_progress_bar():
                total_iterations = shape
                loop_task_id = progress.add_task(f"Looping over {label}", total=total_iterations)
                task_ids[block.uuid] = loop_task_id  # Store the task ID associated with this loop block

                # Track the index for this loop
                loop_indices[block] = 0

                return loop_task_id

            loop_operations.append(create_progress_bar)

            def advance_progress_bar(variable_value: tuple[int | float, ...]) -> None:
                loop_task_id = task_ids[block.uuid]
                progress.update(
                    loop_task_id,
                    description=f"Looping over {label}: {variable_value[0] if len(variable_value) == 1 else variable_value}",
                )
                progress.advance(loop_task_id)

            def advance_loop_index() -> None:
                # Update the loop index
                loop_indices[block] += 1

            uuids = [variable.uuid for variable in self._variables_per_block[block]]
            values = [variable.values for variable in self._variables_per_block[block]]

            for current_values in zip(*values):
                for uuid, value in zip(uuids, current_values):
                    current_value_of_variable[uuid] = value

                loop_operations.append(lambda value=current_values: advance_progress_bar(value))  # type: ignore

                # Process elements within the loop
                loop_operations.extend(process_elements(block.elements))

                loop_operations.append(advance_loop_index)

            def remove_progress_bar():
                progress.remove_task(task_ids[block.uuid])
                del loop_indices[block]

            loop_operations.append(remove_progress_bar)

            return loop_operations

        def process_elements(elements: list[Block | Operation]) -> list[Callable]:
            """Process the elements in a block and store the corresponding operations."""
            elements_operations: list[Callable] = []

            for element in elements:
                if isinstance(element, GetParameter):
                    # Set current value of the variable to None
                    current_value_of_variable[element.variable.uuid] = None  # type: ignore[assignment]
                    # Append a lambda that will call the `platform.get_parameter` method and assign the returned value to the variable
                    elements_operations.append(
                        lambda operation=element: current_value_of_variable.update(
                            {
                                operation.variable.uuid: self.platform.get_parameter(
                                    alias=operation.alias,
                                    parameter=operation.parameter,
                                    channel_id=operation.channel_id,
                                )
                            }
                        )
                    )
                if isinstance(element, SetCrosstalk):
                    elements_operations.append(
                        lambda operation=element: self.platform.set_crosstalk(crosstalk=operation.crosstalk)
                    )
                if isinstance(element, SetParameter):
                    # Append a lambda that will call the `platform.set_parameter` method
                    if isinstance(element.value, Variable):
                        if current_value_of_variable[element.value.uuid] is None:
                            # Variable has no value and it will get it from a `GetOperation` in the future. Thus, don't bind `value` in lambda.
                            elements_operations.append(
                                lambda operation=element: self.platform.set_parameter(
                                    alias=operation.alias,
                                    parameter=operation.parameter,
                                    value=current_value_of_variable[operation.value.uuid],
                                    channel_id=operation.channel_id,
                                )
                            )
                        else:
                            # Variable has a value that was set from a loop. Thus, bind `value` in lambda with the current value of the variable.
                            elements_operations.append(
                                lambda operation=element,
                                value=current_value_of_variable[element.value.uuid]: self.platform.set_parameter(
                                    alias=operation.alias,
                                    parameter=operation.parameter,
                                    value=value,
                                    channel_id=operation.channel_id,
                                )
                            )
                    else:
                        # Value is not a variable. Treat it as a normal Python type.
                        elements_operations.append(
                            lambda operation=element: self.platform.set_parameter(
                                alias=operation.alias,
                                parameter=operation.parameter,
                                value=operation.value,
                                channel_id=operation.channel_id,
                            )
                        )

                if isinstance(element, ExecuteQProgram):
                    qprogram_index = self._qprogram_execution_indices[element]
                    if isinstance(element.qprogram, LambdaType):
                        signature = inspect.signature(element.qprogram)
                        call_parameters: dict[str, int | float] = {}
                        deferred_parameters: dict[str, UUID] = {}

                        # Iterate through parameters and separate the ones that have values and the ones that don't
                        for param in signature.parameters.values():
                            if isinstance(param.default, Variable):
                                variable_value = current_value_of_variable.get(param.default.uuid, None)
                                if variable_value is None:
                                    # The variable doesn't have a value yet; defer binding
                                    deferred_parameters[param.name] = param.default.uuid
                                    # Make sure key exists
                                    current_value_of_variable[param.default.uuid] = None  # type: ignore[assignment]
                                else:
                                    # The variable has a current value; bind it immediately
                                    call_parameters[param.name] = variable_value

                        # Bind the values for known variables, and retrieve deferred ones when the lambda is executed
                        elements_operations.append(
                            lambda operation=element,
                            call_parameters=call_parameters,
                            qprogram_index=qprogram_index: store_results(
                                self.platform.execute_qprogram(
                                    qprogram=operation.qprogram(
                                        **{
                                            **call_parameters,  # Bind the values that are known
                                            **{
                                                param_name: current_value_of_variable[uuid]
                                                for param_name, uuid in deferred_parameters.items()
                                            },  # Defer retrieving missing values
                                        }
                                    ),  # type: ignore
                                    bus_mapping=operation.bus_mapping,
                                    calibration=operation.calibration,
                                    debug=operation.debug,
                                ),
                                qprogram_index,
                            )
                        )
                    else:
                        # Append a lambda that will call the `platform.execute_qprogram` method
                        elements_operations.append(
                            lambda operation=element, qprogram_index=qprogram_index: store_results(
                                self.platform.execute_qprogram(
                                    qprogram=operation.qprogram,
                                    bus_mapping=operation.bus_mapping,
                                    calibration=operation.calibration,
                                    debug=operation.debug,
                                ),
                                qprogram_index,
                            )
                        )
                elif isinstance(element, Block):
                    # Recursively handle elements of the block
                    nested_operations = self._prepare_operations(element, progress)
                    elements_operations.extend(nested_operations)

            return elements_operations

        def store_results(qprogram_results: QProgramResults, qprogram_index: int):
            """Store the result in the correct location within the ExperimentResultsWriter."""
            # Determine the index based on current loop indices and store the results in the ExperimentResultsWriter
            for measurement_index, measurement_result in enumerate(qprogram_results.timeline):
                indices = (qprogram_index, measurement_index, *tuple(index for _, index in loop_indices.items()))
                self._results_writer[indices] = measurement_result.array.T  # type: ignore

        if isinstance(block, (Loop, ForLoop, Parallel)):
            # Handle loops
            operations.extend(handle_loop(block))
        else:
            # Handle generic blocks
            operations.extend(process_elements(block.elements))

        return operations

    def _execute_operations(self, operations: list[Callable], progress: Progress):
        """Run the stored operations in sequence, updating the progress bar."""
        main_task_id = progress.add_task("Executing experiment", total=len(operations))

        for operation in operations:
            # Execute the stored operation and update the main progress bar
            operation()
            progress.advance(main_task_id)

        progress.update(main_task_id, description="Executing experiment (done)")
        progress.refresh()  # Ensure the final state of the progress bar is rendered

    def _inclusive_range(self, start: int | float, stop: int | float, step: int | float) -> np.ndarray:
        # Check if all inputs are integers
        if all(isinstance(x, int) for x in [start, stop, step]):
            # Use numpy.arange for integer ranges
            return np.arange(start, stop + step, step)

        # Define the number of decimal places based on the precision of the step
        decimal_places = -int(np.floor(np.log10(step))) if step < 1 else 0

        # Calculate the number of steps
        num_steps = round((stop - start) / step) + 1

        # Use linspace and then round to avoid floating-point inaccuracies
        result = np.linspace(start, stop, num_steps)
        return np.around(result, decimals=decimal_places)

    def _get_variables_of_loop(self, block: Loop | ForLoop | Parallel) -> list[VariableInfo]:
        variables: dict[UUID, VariableInfo] = {}

        if isinstance(block, (ForLoop, Loop)):
            values = (
                self._inclusive_range(block.start, block.stop, block.step)
                if isinstance(block, ForLoop)
                else block.values
            )
            variable = VariableInfo(uuid=block.variable.uuid, label=block.variable.label, values=values)
            variables[block.variable.uuid] = variable
        else:
            for loop in block.loops:
                values = (
                    self._inclusive_range(loop.start, loop.stop, loop.step)
                    if isinstance(loop, ForLoop)
                    else loop.values
                )
                variable = VariableInfo(uuid=loop.variable.uuid, label=loop.variable.label, values=values)
                variables[loop.variable.uuid] = variable

        self._variables_per_block[block] = list(variables.values())

        # Update all_variables registry
        for variable in variables.values():
            if self._all_variables[variable.uuid]["label"] is None:
                self._all_variables[variable.uuid]["label"] = variable.label
            self._all_variables[variable.uuid]["values"][block.uuid] = variable.values

        return list(variables.values())

    def _create_results_path(self, executed_at: datetime):
        # Get base path and path format from platform
        base_path = self.platform.experiment_results_base_path
        path_format = self.platform.experiment_results_path_format

        # Format date and time for directory names
        date = executed_at.strftime("%Y%m%d")
        timestamp = executed_at.strftime("%H%M%S")
        label = self.experiment.label

        # Format the path based on the path's format
        path = path_format.format(date=date, time=timestamp, label=label)

        # Construct the full path
        path = os.path.join(base_path, path)

        # Ensure it is an absolute path
        path = os.path.abspath(path)

        # Create the directories if they don't exist
        os.makedirs(os.path.dirname(path), exist_ok=True)

        return path

    def _measure_execution_time(self, execution_completed: threading.Event):
        """Measures the execution time while waiting for the experiment to finish."""
        # Start measuring execution time
        start_time = perf_counter()

        # Wait for the experiment to finish
        execution_completed.wait()

        # Stop measuring execution time
        end_time = perf_counter()

        # Return the execution time
        return end_time - start_time

    def execute(self) -> str:
        """
        Executes the experiment and streams the results in real-time.

        This method prepares the experiment by calculating the shape and values of the loops,
        creates callable operations, initializes an ExperimentResultsWriter for real-time result storage,
        and then runs the operations while updating a progress bar.

        Returns:
            str: The path to the file where the results are stored.
        """
        executed_at = datetime.now()

        # Create file path to store results
        results_path = self._create_results_path(executed_at=executed_at)

        # Prepare the results metadata
        self._prepare_metadata(executed_at=executed_at)

        # Create the ExperimentResultsWriter for storing results
        self._results_writer = ExperimentResultsWriter(path=results_path, metadata=self._metadata)

        # Event to signal that the execution has completed
        execution_completed = threading.Event()

        with ThreadPoolExecutor() as executor:
            # Start the _measure_execution_time in a separate thread
            execution_time_future = executor.submit(self._measure_execution_time, execution_completed)

            with self._results_writer:
                with Progress(
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(bar_width=None),
                    "[progress.percentage]{task.percentage:>3.1f}%",
                    TimeElapsedColumn(),
                ) as progress:
                    operations = self._prepare_operations(self.experiment.body, progress)
                    self._execute_operations(operations, progress)

                # Signal that the execution has completed
                execution_completed.set()

                # Retrieve the execution time from the Future
                execution_time = execution_time_future.result()

                # Now write the execution time to the results writer
                self._results_writer.execution_time = execution_time

        return results_path
